from __future__ import annotations

from typing import Dict

import numpy as np

from gpaw.typing import Array3D

_gaunt: Dict[int, np.ndarray] = {}
_nabla: Dict[int, np.ndarray] = {}


def gaunt(lmax: int = 2) -> Array3D:
    r"""Gaunt coefficients

    :::

         ^      ^     -- L      ^
      Y (r)  Y (r) =  > G    Y (r)
       L      L       -- L L  L
        1      2      L   1 2
    """

    if lmax in _gaunt:
        return _gaunt[lmax]

    Lmax = (lmax + 1)**2
    L2max = (2 * lmax + 1)**2

    from gpaw.spherical_harmonics import YL, gam
    G_LLL = np.zeros((Lmax, L2max, L2max))
    for L1 in range(Lmax):
        for L2 in range(L2max):
            for L in range(L2max):
                r = 0.0
                for c1, n1 in YL[L1]:
                    for c2, n2 in YL[L2]:
                        for c, n in YL[L]:
                            nx = n1[0] + n2[0] + n[0]
                            ny = n1[1] + n2[1] + n[1]
                            nz = n1[2] + n2[2] + n[2]
                            r += c * c1 * c2 * gam(nx, ny, nz)
                G_LLL[L1, L2, L] = r
    _gaunt[lmax] = G_LLL
    return G_LLL


def nabla(lmax: int = 2) -> Array3D:
    """Create the array of derivative intergrals.

    :::

      /  ^    ^   1-l' d   l'    ^
      | dr Y (r) r     --[r  Y  (r)]
      /     L           _     L'
                       dr
    """

    if lmax in _nabla:
        return _nabla[lmax]

    Lmax = (lmax + 1)**2
    from gpaw.spherical_harmonics import YL, gam
    Y_LLv = np.zeros((Lmax, Lmax, 3))
    # Insert new values
    for L1 in range(Lmax):
        for L2 in range(Lmax):
            for v in range(3):
                r = 0.0
                for c1, n1 in YL[L1]:
                    for c2, n2 in YL[L2]:
                        n = [0, 0, 0]
                        n[0] = n1[0] + n2[0]
                        n[1] = n1[1] + n2[1]
                        n[2] = n1[2] + n2[2]
                        if n2[v] > 0:
                            # apply derivative
                            n[v] -= 1
                            # add integral
                            r += n2[v] * c1 * c2 * gam(n[0], n[1], n[2])
                Y_LLv[L1, L2, v] = r
    _nabla[lmax] = Y_LLv
    return Y_LLv



from typing import Tuple

from .wrapper import HybridXC

__all__ = ['HybridXC']


def parse_name(name: str) -> Tuple[str, float, float]:
    if name == 'EXX':
        return 'null', 1.0, 0.0
    if name == 'PBE0':
        return 'HYB_GGA_XC_PBEH', 0.25, 0.0
    if name == 'HSE03':
        return 'HYB_GGA_XC_HSE03', 0.25, 0.106
    if name == 'HSE06':
        return 'HYB_GGA_XC_HSE06', 0.25, 0.11
    if name == 'B3LYP':
        return 'HYB_GGA_XC_B3LYP', 0.2, 0.0
    assert False



from time import time

import pytest
from ase import Atoms
from ase.io import read
from ase.optimize import BFGS

from gpaw import GPAW
from gpaw.mpi import world
from gpaw.test import equal


@pytest.mark.later
def test_generic_relax(in_tmp_dir):
    a = 4.0    # Size of unit cell (Angstrom)
    c = a / 2
    d = 0.74  # Experimental bond length
    molecule = Atoms('H2',
                     [(c - d / 2, c, c),
                      (c + d / 2, c, c)],
                     cell=(a, a, a),
                     pbc=False)
    calc = GPAW(h=0.2, nbands=1, xc={'name': 'PBE', 'stencil': 1}, txt=None,
                poissonsolver={'name': 'fd'})
    molecule.calc = calc
    e1 = molecule.get_potential_energy()
    calc.write('H2.gpw')
    calc.write('H2a.gpw', mode='all')
    molecule.get_forces()
    calc.write('H2f.gpw')
    calc.write('H2fa.gpw', mode='all')

    def timer(func, *args, **kwargs):
        t0 = time()
        ret = func(*args, **kwargs)
        return ret, time() - t0

    molecule = GPAW('H2.gpw', txt=None).get_atoms()
    f1, t1 = timer(molecule.get_forces)
    molecule = GPAW('H2a.gpw', txt=None).get_atoms()
    f2, t2 = timer(molecule.get_forces)
    molecule = GPAW('H2f.gpw', txt=None).get_atoms()
    f3, t3 = timer(molecule.get_forces)
    molecule = GPAW('H2fa.gpw', txt=None).get_atoms()
    f4, t4 = timer(molecule.get_forces)
    print('timing:', t1, t2, t3, t4)
    assert t2 < 0.6 * t1
    assert t3 < 0.5
    assert t4 < 0.5
    print(f1)
    print(f2)
    print(f3)
    print(f4)
    assert sum((f1 - f4).ravel()**2) < 1e-6
    assert sum((f2 - f4).ravel()**2) < 1e-6
    assert sum((f3 - f4).ravel()**2) < 1e-6

    positions = molecule.get_positions()
    #                 x-coordinate      x-coordinate
    #                 v                 v
    d0 = positions[1, 0] - positions[0, 0]
    #              ^                 ^
    #              second atom       first atom

    print('experimental bond length:')
    print('hydrogen molecule energy: %7.3f eV' % e1)
    print('bondlength              : %7.3f Ang' % d0)

    # Find the theoretical bond length:
    relax = BFGS(molecule)
    relax.run(fmax=0.05)

    e2 = molecule.get_potential_energy()

    positions = molecule.get_positions()
    #                 x-coordinate      x-coordinate
    #                 v                 v
    d0 = positions[1, 0] - positions[0, 0]
    #              ^                 ^
    #              second atom       first atom

    print('PBE energy minimum:')
    print('hydrogen molecule energy: %7.3f eV' % e2)
    print('bondlength              : %7.3f Ang' % d0)

    molecule = GPAW('H2fa.gpw', txt='H2.txt').get_atoms()
    relax = BFGS(molecule)
    relax.run(fmax=0.05)
    e2q = molecule.get_potential_energy()
    positions = molecule.get_positions()
    d0q = positions[1, 0] - positions[0, 0]
    assert abs(e2 - e2q) < 2e-6
    assert abs(d0q - d0) < 4e-4

    f0 = molecule.get_forces()
    del relax, molecule

    world.barrier()  # syncronize before reading text output file
    f = read('H2.txt').get_forces()
    assert abs(f - f0).max() < 5e-6  # 5 digits in txt file

    energy_tolerance = 0.002
    equal(e1, -6.287873, energy_tolerance)
    equal(e2, -6.290744, energy_tolerance)
    equal(e2q, -6.290744, energy_tolerance)



from math import sqrt, pi
import numpy as np

from gpaw.utilities import pack, unpack2
from gpaw.utilities.tools import pick
from gpaw.lfc import LocalizedFunctionsCollection as LFC, BasisFunctions


# XXX Document what is the difference between PairDensity2 and 1.
class PairDensity2:
    def __init__(self, density, spos_ac, finegrid):
        """Initialization needs a paw instance, and whether the compensated
        pair density should be on the fine grid (boolean)"""

        self.density = density
        self.finegrid = finegrid

        if not finegrid:
            density.Ghat = LFC(density.gd,
                               [setup.ghat_l
                                for setup in density.setups],
                               integral=sqrt(4 * pi))
            density.Ghat.set_positions(spos_ac)

    def initialize(self, kpt, n1, n2):
        """Set wave function indices."""
        self.n1 = n1
        self.n2 = n2
        self.spin = kpt.s
        self.P_ani = kpt.P_ani
        self.psit1_G = pick(kpt.psit_nG, n1)
        self.psit2_G = pick(kpt.psit_nG, n2)

    def get_coarse(self, nt_G):
        """Get pair density"""
        np.multiply(self.psit1_G.conj(), self.psit2_G, nt_G)

    def add_compensation_charges(self, nt_G, rhot_g):
        """Add compensation charges to input pair density, which
        is interpolated to the fine grid if needed."""

        if self.finegrid:
            # interpolate the pair density to the fine grid
            self.density.interpolator.apply(nt_G, rhot_g)
        else:
            # copy values
            rhot_g[:] = nt_G

        # Determine the compensation charges for each nucleus
        Q_aL = {}
        for a, P_ni in self.P_ani.items():
            assert P_ni.dtype == float
            # Generate density matrix
            P1_i = P_ni[self.n1]
            P2_i = P_ni[self.n2]
            D_ii = np.outer(P1_i.conj(), P2_i)
            # allowed to pack as used in the scalar product with
            # the symmetric array Delta_pL
            D_p = pack(D_ii, tolerance=1e30)

            # Determine compensation charge coefficients:
            Q_aL[a] = np.dot(D_p, self.density.setups[a].Delta_pL)

        # Add compensation charges
        if self.finegrid:
            self.density.ghat.add(rhot_g, Q_aL)
        else:
            self.density.Ghat.add(rhot_g, Q_aL)


class PairDensity:
    def __init__(self, paw):
        self.set_paw(paw)

    def set_paw(self, paw):
        """basic initialisation knowing the calculator"""
        self.wfs = paw.wfs
        self.lcao = paw.wfs.mode == 'lcao'
        self.density = paw.density
        self.setups = paw.wfs.setups
        self.spos_ac = paw.spos_ac

        self.spin = 0
        self.k = 0
        self.weight = 0.0

        if self.lcao:
            assert paw.wfs.dtype == float
            self.wfs = paw.wfs

    def initialize(self, kpt, i, j):
        """initialize yourself with the wavefunctions"""
        self.i = i
        self.j = j

        if kpt is not None:
            self.spin = kpt.s
            self.k = kpt.k
            self.weight = kpt.weight
            self.P_ani = kpt.P_ani
            if self.lcao:
                self.q = kpt.q
                self.wfi_M = kpt.C_nM[i]
                self.wfj_M = kpt.C_nM[j]
            else:
                self.wfi = kpt.psit_nG[i]
                self.wfj = kpt.psit_nG[j]

    def get_lcao(self, finegrid=False):
        """Get pair density"""
        # Expand the pair density as density matrix
        rho_MM = (0.5 * np.outer(self.wfi_M, self.wfj_M) +
                  0.5 * np.outer(self.wfj_M, self.wfi_M))

        rho_G = self.density.gd.zeros()
        self.wfs.basis_functions.construct_density(rho_MM, rho_G, self.q)

        if not finegrid:
            return rho_G

        # interpolate the pair density to the fine grid
        rho_g = self.density.finegd.zeros()
        self.density.interpolator.apply(rho_G, rho_g)

        return rho_g

    def get(self, finegrid=False):
        """Get pair density"""

        if self.lcao:
            return self.get_lcao(finegrid)

        nijt_G = self.wfi.conj() * self.wfj
        if not finegrid:
            return nijt_G

        # interpolate the pair density to the fine grid
        nijt_g = self.density.finegd.empty(dtype=nijt_G.dtype)
        self.density.interpolator.apply(nijt_G, nijt_g)

        return nijt_g

    def with_compensation_charges(self, finegrid=False):
        """Get pair density including the compensation charges"""
        rhot_g = self.get(finegrid)

        # Determine the compensation charges for each nucleus
        Q_aL = {}
        for a, P_ni in self.P_ani.items():
            # Generate density matrix
            Pi_i = P_ni[self.i].conj()
            Pj_i = P_ni[self.j]
            D_ii = np.outer(Pi_i, Pj_i)
            # allowed to pack as used in the scalar product with
            # the symmetric array Delta_pL
            D_p = pack(D_ii)

            # Determine compensation charge coefficients:
            Q_aL[a] = np.dot(D_p, self.setups[a].Delta_pL)

        # Add compensation charges
        if finegrid:
            self.density.ghat.add(rhot_g, Q_aL)
        else:
            if not hasattr(self.density, 'Ghat'):
                self.density.Ghat = LFC(self.density.gd,
                                        [setup.ghat_l
                                         for setup in self.setups],
                                        integral=sqrt(4 * pi))
                self.density.Ghat.set_positions(self.spos_ac)
            self.density.Ghat.add(rhot_g, Q_aL)

        return rhot_g

    def with_ae_corrections(self, finegrid=False):
        """Get pair density including the AE corrections"""
        nij_g = self.get(finegrid)

        # Generate the density matrix
        D_ap = {}
#        D_aii = {}
        for a, P_ni in self.P_ani.items():
            Pi_i = P_ni[self.i]
            Pj_i = P_ni[self.j]
            D_ii = np.outer(Pi_i.conj(), Pj_i)
            # Note: D_ii is not symmetric but the products of partial waves are
            # so that we can pack
            D_ap[a] = pack(D_ii)
#            D_aii[a] = D_ii

        # Load partial waves if needed
        if ((finegrid and (not hasattr(self, 'phi'))) or
            ((not finegrid) and (not hasattr(self, 'Phi')))):

            # Splines
            splines = {}
            phi_aj = []
            phit_aj = []
            for a, id in enumerate(self.setups.id_a):
                if id in splines:
                    phi_j, phit_j = splines[id]
                else:
                    # Load splines:
                    phi_j, phit_j = self.setups[a].get_partial_waves()[:2]
                    splines[id] = (phi_j, phit_j)
                phi_aj.append(phi_j)
                phit_aj.append(phit_j)

            # Store partial waves as class variables
            if finegrid:
                gd = self.density.finegd
                self.__class__.phi = BasisFunctions(gd, phi_aj)
                self.__class__.phit = BasisFunctions(gd, phit_aj)
                self.__class__.phi.set_positions(self.spos_ac)
                self.__class__.phit.set_positions(self.spos_ac)
            else:
                gd = self.density.gd
                self.__class__.Phi = BasisFunctions(gd, phi_aj)
                self.__class__.Phit = BasisFunctions(gd, phit_aj)
                self.__class__.Phi.set_positions(self.spos_ac)
                self.__class__.Phit.set_positions(self.spos_ac)

        # Add AE corrections
        if finegrid:
            phi = self.phi
            phit = self.phit
            gd = self.density.finegd
        else:
            phi = self.Phi
            phit = self.Phit
            gd = self.density.gd

        rho_MM = np.zeros((phi.Mmax, phi.Mmax))
        M1 = 0
        for a, setup in enumerate(self.setups):
            ni = setup.ni
            D_p = D_ap.get(a)
            if D_p is None:
                D_p = np.empty((ni * (ni + 1) // 2))
            if gd.comm.size > 1:
                gd.comm.broadcast(D_p, self.wfs.partition.rank_a[a])
            D_ii = unpack2(D_p)
#            D_ii = D_aii.get(a)
#            if D_ii is None:
#                D_ii = np.empty((ni, ni))
#            if gd.comm.size > 1:
#                gd.comm.broadcast(D_ii, self.wfs.atom_partition.rank_a[a])
            M2 = M1 + ni
            rho_MM[M1:M2, M1:M2] = D_ii
            M1 = M2

        # construct_density assumes symmetric rho_MM and
        # takes only the upper half of it
        phi.construct_density(rho_MM, nij_g, q=-1)
        phit.construct_density(-rho_MM, nij_g, q=-1)
        # TODO: use ae_valence_density_correction
#        phi.lfc.ae_valence_density_correction(
#            rho_MM, nij_g, np.zeros(len(phi.M_W), np.intc), np.zeros(self.na))
#        phit.lfc.ae_valence_density_correction(
#            -rho_MM, nij_g, np.zeros(len(phit.M_W), np.intc),
#            np.zeros(self.na))

        return nij_g



from ase import Atom, Atoms
from gpaw import GPAW
from gpaw.test import equal
import numpy as np


def test_generic_bulk():
    bulk = Atoms([Atom('Li')], pbc=True)
    k = 4
    g = 8
    calc = GPAW(gpts=(g, g, g), kpts=(k, k, k), nbands=2)
    bulk.calc = calc
    a = np.linspace(2.6, 2.8, 5)
    e = []
    for x in a:
        bulk.set_cell((x, x, x))
        e1 = bulk.get_potential_energy()
        e.append(e1)

    fit = np.polyfit(a, e, 2)
    a0 = np.roots(np.polyder(fit, 1))[0]
    e0 = np.polyval(fit, a0)
    print('a,e =', a0, e0)
    equal(a0, 2.641, 0.001)
    equal(e0, -1.98357, 0.0002)

    energy_tolerance = 0.0002
    equal(e1, -1.96157, energy_tolerance)



import pytest
import numpy as np
from gpaw.xc.fxc_kernels import get_pbe_fxc_and_intermediate_derivatives


def eps_unif_x(kf_g):
    return -3 * kf_g / (4.0 * np.pi)


# (We test "small" and "large" densities separately)
@pytest.mark.parametrize(
    'n_g', [
        np.linspace(0.001, 0.02, 10000),
        np.linspace(0.02, 0.5, 1000)
    ])
def test_pbe_x_derivs(n_g):
    # Here we test d(fxc(n))/dn.  We create a linspace of n and
    # calculate all the derivatives.  However we put a dummy (constant)
    # value for s2_g (the gradient) which makes the numerics sensitive
    # (I think) for small densities.
    #
    # This test guards against the problem observed in #723.
    dn = n_g[1] - n_g[0]
    # (assuming np.linspace)

    gradn2 = np.ones_like(n_g)
    kf_g = (3. * np.pi**2 * n_g)**(1 / 3.)
    s2_g = gradn2 / (4 * kf_g**2 * n_g**2)

    fxc_g, F_g, Fn_g, Fnn_g = get_pbe_fxc_and_intermediate_derivatives(
        n_g, s2_g)

    def numderiv(y_g):
        return (y_g[1:] - y_g[:-1]) / dn

    E_g = n_g * eps_unif_x(kf_g) * F_g
    d2Edn2_num = numderiv(numderiv(E_g))

    assert fxc_g[1:-1] == pytest.approx(d2Edn2_num, abs=1e-4, rel=1e-4)
    # We could also test the other derivatives (Fn, Fnn)
    # but that's probably not highest priority.



import pytest

from ase.build import bulk

from gpaw import GPAW

from gpaw.elph import DisplacementRunner
from gpaw.elph import Supercell

SUPERCELL = (2, 1, 1)


def get_calc(txt, parallel={}):
    return GPAW(mode='lcao',
                basis='sz(dzp)',
                kpts={'size': (1, 2, 2), 'gamma': False},
                symmetry={'point_group': False},
                convergence={'density': 1.5e-1, 'energy': 1},
                parallel=parallel,
                txt=txt)


# NOTE: This fixture might need a proper scope assigned
@pytest.fixture(scope='module')
def elph_cache(module_tmp_path):
    """Minimum elph cache for Li

    Uses 1x2x2 k-points and 2x1x1 SC to allow for parallelisaiton
    test.
    Takes 6s on 4 cores.
    """
    atoms = bulk('Li', crystalstructure='bcc', a=3.51, cubic=True)
    calc = get_calc(txt='elph_li.txt')
    atoms.calc = calc
    elph = DisplacementRunner(atoms, calc,
                              supercell=SUPERCELL, name='elph',
                              calculate_forces=True)
    elph.run()
    return elph


# NOTE: This fixture might need a proper scope assigned
@pytest.fixture(scope='module')
def supercell_cache(module_tmp_path, elph_cache):
    atoms = bulk('Li', crystalstructure='bcc', a=3.51, cubic=True)
    atoms_N = atoms * SUPERCELL
    elph_cache
    calc = get_calc(parallel={'domain': 1, 'band': 1},
                    # parallel={'sl_auto': True, 'augment_grids':True,
                    #          'band': 2, 'kpt': 1, 'domain': 1 },
                    txt='gs_li.txt')
    atoms_N.calc = calc
    atoms_N.get_potential_energy()

    # create supercell cache
    sc = Supercell(atoms, supercell=SUPERCELL)
    sc.calculate_supercell_matrix(calc)
    return sc



"""This script asserts that the chi's obtained
from GS calculations using symmetries
and GS calculations not using symmetries return
the same results. Tests that the chi's are element-wise
equal to a tolerance of 1e-10.
"""

import pytest
import numpy as np

from gpaw.response.chi0 import Chi0


@pytest.mark.response
@pytest.mark.slow
def test_response_symmetry(gpw_files):
    data_s = []
    for name in ['ti2o4_pw_nosym_wfs', 'ti2o4_pw_wfs']:
        X = Chi0(gpw_files[name])
        chi_t = X.calculate([1. / 4, 0, 0])
        data_s.append((chi_t.chi0_WgG, chi_t.chi0_WxvG, chi_t.chi0_Wvv))

    msg = 'Difference in Chi when turning off symmetries!'

    while len(data_s):
        data1 = data_s.pop()
        for data2 in data_s:
            for dat1, dat2 in zip(data1, data2):
                if dat1 is not None:
                    assert (np.abs(dat1 - dat2).max() ==
                            pytest.approx(0, abs=0.001)), msg



import pytest
from gpaw.mpi import world
import numpy as np
from gpaw.response.g0w0 import G0W0
import pickle


@pytest.mark.response
def test_do_GW_too(in_tmp_dir, gpw_files, scalapack, needs_ase_master):
    gw0 = G0W0(gpw_files['c_pw_wfs'], 'gw0',
               bands=(3, 5),
               nblocks=1,
               ecut_extrapolation=True,
               ecut=40)
    results0 = gw0.calculate()

    gw = G0W0(gpw_files['c_pw_wfs'], 'gwtoo',
              bands=(3, 5),
              nblocks=1,
              xc='rALDA',
              ecut_extrapolation=True,
              ecut=40,
              fxc_mode='GWP',
              do_GW_too=True)

    gw.calculate()

    world.barrier()

    files = gw.savepckl()

    with open(files['GW'], 'rb') as handle:
        results_GW = pickle.load(handle)
    np.testing.assert_allclose(results0['qp'], results_GW['qp'], rtol=1e-03)



import pytest
from ase import Atoms
from gpaw.mpi import world
from gpaw import GPAW, PW
from gpaw.response.df import DielectricFunction
from gpaw.test import findpeak

# Comparing the EELS spectrum of sodium for different block
# parallelizations. Intended to be run with 8 cores.

pytestmark = pytest.mark.skipif(world.size < 4, reason='world.size < 4')


@pytest.mark.response
@pytest.mark.slow
def test_response_na_plasmons_tetrahedron(in_tmp_dir, scalapack):
    a = 4.23 / 2.0
    a1 = Atoms('Na',
               scaled_positions=[[0, 0, 0]],
               cell=(a, a, a),
               pbc=True)

    a1.calc = GPAW(mode=PW(250),
                   kpts={'size': (10, 10, 10), 'gamma': True},
                   parallel={'band': 1},
                   txt='small.txt')

    a1.get_potential_energy()
    a1.calc.diagonalize_full_hamiltonian(nbands=20)
    a1.calc.write('gs_Na.gpw', 'all')

    kwargs = {'integrationmode': 'tetrahedron integration',
              'ecut': 40}

    # Calculate the dielectric functions: tetrahedral 1 block
    df1 = DielectricFunction('gs_Na.gpw',
                             rate=0.001,
                             nblocks=1,
                             txt='1block.txt',
                             **kwargs)
    df1NLFCx, df1LFCx = df1.get_dielectric_function(direction='x')

    # tetrahedral 4 blocks
    df2 = DielectricFunction('gs_Na.gpw',
                             rate=0.001,
                             nblocks=4,
                             txt='4block.txt',
                             **kwargs)
    df2NLFCx, df2LFCx = df2.get_dielectric_function(direction='x')

    # tetrahedron integration 4 blocks with large eta
    kwargs.update({'eta': 4.25})
    df3 = DielectricFunction('gs_Na.gpw',
                             rate=0.001,
                             nblocks=4,
                             txt='4block.txt',
                             **kwargs)
    df3NLFCx, df3LFCx = df3.get_dielectric_function(direction='x')

    # point integration 4 blocks with large eta (smearing)
    kwargs.update({'integrationmode': None, 'eta': 4.25})
    df4 = DielectricFunction('gs_Na.gpw',
                             rate=0.001,
                             nblocks=4,
                             txt='4block.txt',
                             **kwargs)
    df4NLFCx, df4LFCx = df4.get_dielectric_function(direction='x')

    # Compare plasmon frequencies and intensities
    w_w = df1.wd.omega_w

    w1, I1 = findpeak(w_w, -(1. / df1LFCx).imag)
    w2, I2 = findpeak(w_w, -(1. / df2LFCx).imag)
    w3, I3 = findpeak(w_w, -(1. / df3LFCx).imag)
    w4, I4 = findpeak(w_w, -(1. / df4LFCx).imag)

    # omegas don't change
    assert [w3, w4] == pytest.approx([0.283057, 0.300520], 1e-2)
    assert w1 == pytest.approx(w2, 1e-2)  # omega: serial vs parallel
    # omega: PI & TI w/ large eta are aprx equal
    assert w4 == pytest.approx(w3, 2e-2, abs=True)
    assert I1 == pytest.approx(I2, 1e-3)  # intensity: serial vs parallel



import pytest
import numpy as np
from ase.parallel import parprint
from gpaw import GPAW
from gpaw.mpi import world


def numeric_stress(atoms, d=1e-6, component=None):
    cell = atoms.cell.copy()
    V = atoms.get_volume()
    for i in range(3):
        x = np.eye(3)
        if component == (i, i):
            x[i, i] += d
            atoms.set_cell(np.dot(cell, x), scale_atoms=True)
            eplus = atoms.get_potential_energy(force_consistent=True)

            x[i, i] -= 2 * d
            atoms.set_cell(np.dot(cell, x), scale_atoms=True)
            eminus = atoms.get_potential_energy(force_consistent=True)
            atoms.set_cell(cell, scale_atoms=True)
            return (eplus - eminus) / (2 * d * V)

        if (component == (i, (i - 2) % 3)) or (component == ((i - 2) % 3, i)):
            j = i - 2
            x[i, j] = d
            x[j, i] = d
            atoms.set_cell(np.dot(cell, x), scale_atoms=True)
            eplus = atoms.get_potential_energy(force_consistent=True)

            x[i, j] = -d
            x[j, i] = -d
            atoms.set_cell(np.dot(cell, x), scale_atoms=True)
            eminus = atoms.get_potential_energy(force_consistent=True)
            atoms.set_cell(cell, scale_atoms=True)
            return (eplus - eminus) / (4 * d * V)

    raise ValueError(f'Invalid component {component}')


@pytest.mark.later
@pytest.mark.skipif(world.size >= 4, reason='See #898')
def test_xc_qna_stress(in_tmp_dir, gpw_files):
    calc = GPAW(gpw_files['Cu3Au_qna'])
    atoms = calc.get_atoms()
    atoms.set_cell(np.dot(atoms.cell,
                          [[1.02, 0, 0.03],
                           [0, 0.99, -0.02],
                           [0.2, -0.01, 1.03]]),
                   scale_atoms=True)

    s_analytical = atoms.get_stress(voigt=False)
    print(s_analytical)
    components = [(0, 0), (0, 1), (0, 2), (1, 1), (1, 2), (2, 2)]
    for componentid in [1]:
        component = components[componentid]
        s_numerical = numeric_stress(atoms, 1e-5, component)
        s_err = s_numerical - s_analytical.__getitem__(component)

        parprint('Analytical stress:', s_analytical.__getitem__(component))
        parprint('Numerical stress :', s_numerical)
        parprint('Error in stress  :', s_err)
        assert np.abs(s_err) < 0.002



"""This module implements the calculation of the electron-phonon couplings."""

__all__ = ["ElectronPhononCoupling"]

import numpy as np
import numpy.fft as fft


class ElectronPhononCoupling:
    """..."""
    def __init__(self, atoms, gd, kd, calc=None, dmatrix=None, dtype=float):
        """...

        Parameters
        ----------
        atoms: ...
            Atoms in the unit cell.
        gd: GridDescriptor
            Descriptor for the grid on which the derivatives of the effective
            potential are calculated (most likely the coarse grid in the
            ground-state calculation).
        kd: KPointDescriptor
            Descriptor for the q-vector grid on which the derivatives of the
            effective potential are calculated. This will be the same as the
            one used in the ``PhononCalculator`` to obtain the phonons.
        calc: Calculator
            Calculator for a ground-state calculation.
        dmatrix: DynamicalMatrix
            The dynamical matrix. Used to obtain the phonon polarization
            vectors which appears in the coupling elements.

        """

        self.atoms = atoms
        self.gd = gd
        self.kd = kd
        self.calc = calc
        self.dmatrix = dmatrix
        self.dtype = dtype

        # List for effective-potential derivatives
        self.v1_eff_qavG = []
        self.v1_eff_kavG = None

    def __getstate__(self):
        """Method used to pickle an instance of this object.

        Bound method attributes cannot be pickled and must therefore be deleted
        before an instance is dumped to file.

        """

        # Get state of object and take care of troublesome attributes
        state = dict(self.__dict__)
        state.pop('calc')
        # state['kd'].__dict__['comm'] = serial_comm
        # state['atoms'].__dict__.pop('calc')

        return state

    def collect(self):
        """Collect derivatives from slaves."""

        N = self.atoms.get_number_of_atoms()
        n_c = tuple(self.gd.n_c)
        mynks = self.kd.mynks

        # Collect from slaves
        if self.kd.comm.rank == 0:
            # Global array
            self.v1_eff_kavG = self.gd.empty(n=(self.kd.nibzkpts, N, 3),
                                             dtype=self.dtype)
            v1_eff_qavG = np.asarray(self.v1_eff_qavG)
            v1_eff_qavG.shape = (mynks, N, 3) + n_c
            uslice = self.kd.get_slice()
            self.v1_eff_kavG[uslice] = v1_eff_qavG

            for slave_rank in range(1, self.kd.comm.size):
                uslice = self.kd.get_slice(rank=slave_rank)
                nks = uslice.stop - uslice.start
                v1_eff_qavG = self.gd.empty(n=(nks, N, 3), dtype=self.dtype)
                self.kd.comm.receive(v1_eff_qavG, slave_rank, tag=123)
                self.v1_eff_kavG[uslice] = v1_eff_qavG
        else:
            v1_eff_qavG = np.asarray(self.v1_eff_qavG)
            v1_eff_qavG.shape = (mynks, N, 3) + n_c
            self.kd.comm.send(v1_eff_qavG, 0, tag=123)

    def evaluate_integrals(self, kpts=None, bands=[4, 5]):
        """Calculate matrix elements of the potential derivatives.

        Start by evaluating the integrals between the already calculated
        wave-functions.

        Parameters
        ----------
        kpts: ???
            Specification of the k-point grid.
        bands: list
            List of band indices to include.

        """

        assert self.calc is not None, "Calculator not set."

        # Phonon k-point descriptor
        kd_p = self.kd

        # Find k-points to be included
        if kpts is None:
            # Use existing wave-functions in the calculator
            pass
        elif isinstance(kpts[0], int):
            # List of ints
            self.calc.set(kpts=kpts)
            # assert not smaller than q-grid
        else:
            # Shifted k-point grid (should be possible to make it dense too)
            kpts = np.array(kpts)
            assert kpts.shape == (1, 3), "Invalid kpts value"
            kpts = kd_p.bzk_kc + kpts
            # Translate back into the first BZ
            kpts[np.where(kpts > 0.5)] -= 1.
            kpts[np.where(kpts <= -0.5)] += 1.
            self.calc.set(kpts=kpts)

        # Do a single-shot calculation to get the required wave-functions


#         self.calc.set(fixdensity=True,
#                       nbands=10,
#                       convergence=dict(nbands=8),
#                       # basis='dzp',
#                       eigensolver='cg',
#                       symmetry='off')
#         self.calc.get_potential_energy()

# Electronic k-point descriptor
        kd_e = self.calc.wfs.kd

        # Grid descriptor
        gd = self.calc.wfs.gd

        # Number of matrix elements along the different indices
        n_q = kd_p.nbzkpts
        n_k = kd_e.nbzkpts
        n_a = self.atoms.get_number_of_atoms()
        n_bands = len(bands)
        # Array for matrix elements
        g_qavknn = np.empty((n_q, n_a, 3, n_k, n_bands, n_bands),
                            dtype=self.dtype)

        # Calculate matrix elements
        for q, q_c in enumerate(kd_p.bzk_kc):

            kplusq_k = kd_e.find_k_plus_q(q_c)

            for a, atom in enumerate(self.atoms):

                for v in [0, 1, 2]:

                    v1_eff_G = self.v1_eff_kavG[q, a, v]

                    for k, k_c in enumerate(kd_e.bzk_kc):

                        kplusq = kplusq_k[k]

                        kpt = self.calc.wfs.kpt_u[k]
                        kplusqpt = self.calc.wfs.kpt_u[kplusq]

                        # XXX Tar load
                        psit_k_nG = kpt.psit_nG[:][bands]
                        psit_kplusq_nG = kplusqpt.psit_nG[:][bands]

                        a_n = psit_kplusq_nG.conj() * v1_eff_G

                        for n, psit_G in enumerate(psit_k_nG):

                            g_n = gd.integrate(a_n * psit_G)
                            # XXX Squared elements
                            g_qavknn[q, a, v, k, n] = (g_n * g_n.conj()).real

        return g_qavknn

    def real_space(self, g_qavknn):
        """Fourier transform to real-space."""

        # Shape of q-point grid
        N_c = tuple(self.kd.N_c)

        # Reshape before Fourier transforming
        shape = g_qavknn.shape
        gq_cccavknn = g_qavknn.reshape(N_c + shape[1:])

        gm_cccavknn = fft.ifftn(fft.ifftshift(gq_cccavknn, axes=(0, 1, 2)),
                                axes=(0, 1, 2))

        # Reshape for the evaluation of the fourier sums
        g_mavknn = gm_cccavknn.reshape(shape)

        # Corresponding R_m vectors in units of the lattice vectors
        N1_c = np.array(N_c)[:, np.newaxis]
        R_cm = np.indices(N1_c).reshape(3, -1)
        R_cm += N1_c // 2
        R_cm %= N1_c
        R_cm -= N1_c // 2

        return g_mavknn, R_cm

    def fourier_interpolate(self, g_mavknn, N=2**5):
        """Fourier interpolate the real-space matrix elements."""

        # Reshape before fourier transforming
        shape = g_mavknn.shape
        N_c = tuple(self.kd.N_c)
        gm_cccavknn = g_mavknn.reshape(N_c + shape[1:])
        # Zero pad
        gm_cccavknn_ = np.zeros((N, N, N) + shape[1:], dtype=self.dtype)
        # Insert existing values
        i, j, k = self.kd.N_c / 2 + 1
        gm_cccavknn_[:i, :j, :k] = gm_cccavknn[:i, :j, :k]
        i, j, k = -self.kd.N_c / 2
        gm_cccavknn_[i:, j:, k:] = gm_cccavknn[i:, j:, k:]

        # Fourier transform
        gq_cccavknn = fft.fftn(gm_cccavknn_, axes=(0, 1, 2))

        return gq_cccavknn



from __future__ import annotations

from typing import Callable

import numpy as np
from gpaw.core.atom_arrays import (AtomArrays, AtomArraysLayout,
                                   AtomDistribution)
from gpaw.core.matrix import Matrix
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new import cached_property
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.new.wave_functions import WaveFunctions
from gpaw.setup import Setups
from gpaw.typing import Array2D, Array3D
from gpaw.new.potential import Potential


class LCAOWaveFunctions(WaveFunctions):
    xp = np

    def __init__(self,
                 *,
                 setups: Setups,
                 density_adder: Callable[[Array2D, Array3D], None],
                 tci_derivatives,
                 basis,
                 C_nM: Matrix,
                 S_MM: Matrix,
                 T_MM: Array2D,
                 P_aMi,
                 fracpos_ac: Array2D,
                 atomdist: AtomDistribution,
                 kpt_c=(0.0, 0.0, 0.0),
                 domain_comm: MPIComm = serial_comm,
                 spin: int = 0,
                 q: int = 0,
                 k: int = 0,
                 weight: float = 1.0,
                 ncomponents: int = 1):
        super().__init__(setups=setups,
                         nbands=C_nM.shape[0],
                         spin=spin,
                         q=q,
                         k=k,
                         kpt_c=kpt_c,
                         weight=weight,
                         fracpos_ac=fracpos_ac,
                         atomdist=atomdist,
                         ncomponents=ncomponents,
                         dtype=C_nM.dtype,
                         domain_comm=domain_comm,
                         band_comm=C_nM.dist.comm)
        self.density_adder = density_adder
        self.tci_derivatives = tci_derivatives
        self.basis = basis
        self.C_nM = C_nM
        self.T_MM = T_MM
        self.S_MM = S_MM
        self.P_aMi = P_aMi

        self.bytes_per_band = (self.array_shape(global_shape=True)[0] *
                               C_nM.data.itemsize)

        # This is for TB-mode (and MYPY):
        self.V_MM: Matrix

    @cached_property
    def L_MM(self):
        S_MM = self.S_MM.copy()
        S_MM.invcholesky()
        if self.ncomponents < 4:
            return S_MM
        M, M = S_MM.shape
        L_sMsM = Matrix(2 * M, 2 * M, dtype=complex)
        L_sMsM.data[:] = 0.0
        L_sMsM.data[:M, :M] = S_MM.data
        L_sMsM.data[M:, M:] = S_MM.data
        return L_sMsM

    def _short_string(self, global_shape):
        return f'basis functions: {global_shape[0]}'

    def array_shape(self, global_shape=False):
        if global_shape:
            return self.C_nM.shape[1:]
        1 / 0

    @property
    def P_ani(self):
        if self._P_ani is None:
            atomdist = AtomDistribution.from_atom_indices(
                list(self.P_aMi),
                self.domain_comm,
                natoms=len(self.setups))
            layout = AtomArraysLayout([setup.ni for setup in self.setups],
                                      atomdist=atomdist,
                                      dtype=self.dtype)
            self._P_ani = layout.empty(self.nbands,
                                       comm=self.C_nM.dist.comm)
            for a, P_Mi in self.P_aMi.items():
                self._P_ani[a][:] = (self.C_nM.data @ P_Mi)
        return self._P_ani

    def add_to_density(self,
                       nt_sR,
                       D_asii: AtomArrays) -> None:
        """Add density from wave functions.

        Adds to ``nt_sR`` and ``D_asii``.
        """
        rho_MM = self.calculate_density_matrix()
        self.density_adder(rho_MM, nt_sR.data[self.spin])
        f_n = self.weight * self.spin_degeneracy * self.myocc_n
        self.add_to_atomic_density_matrices(f_n, D_asii)

    def gather_wave_function_coefficients(self) -> np.ndarray:
        C_nM = self.C_nM.gather()
        if C_nM is not None:
            return C_nM.data
        return None

    def calculate_density_matrix(self, eigs=False) -> np.ndarray:
        """Calculate the density matrix.

        The density matrix is:::

                -- *
          ρ   = > C  C   f
           μν   -- nμ nν  n
                n

        Returns
        -------
        The density matrix in the LCAO basis
        """
        if self.domain_comm.rank == 0:
            f_n = self.weight * self.spin_degeneracy * self.myocc_n
            if eigs:
                f_n *= self.myeig_n
            C_nM = self.C_nM.data
            rho_MM = (C_nM.T.conj() * f_n) @ C_nM
            self.band_comm.sum(rho_MM)
        else:
            rho_MM = np.empty_like(self.T_MM)
        self.domain_comm.broadcast(rho_MM, 0)

        return rho_MM

    def to_uniform_grid_wave_functions(self,
                                       grid,
                                       basis):
        grid = grid.new(kpt=self.kpt_c, dtype=self.dtype)
        psit_nR = grid.zeros(self.nbands, self.band_comm)
        basis.lcao_to_grid(self.C_nM.data, psit_nR.data, self.q)

        return PWFDWaveFunctions(
            psit_nR,
            self.spin,
            self.q,
            self.k,
            self.setups,
            self.fracpos_ac,
            self.atomdist,
            self.weight,
            self.ncomponents)

    def collect(self,
                n1: int = 0,
                n2: int = 0) -> LCAOWaveFunctions | None:
        # Quick'n'dirty implementation
        # We should generalize the PW+FD method
        assert self.band_comm.size == 1
        assert self.domain_comm.size == 1
        n2 = n2 or self.nbands + n2
        return LCAOWaveFunctions(
            setups=self.setups,
            density_adder=self.density_adder,
            tci_derivatives=self.tci_derivatives,
            basis=self.basis,
            C_nM=Matrix(n2 - n1,
                        self.C_nM.shape[1],
                        data=self.C_nM.data[n1:n2].copy()),
            S_MM=self.S_MM,
            T_MM=self.T_MM,
            P_aMi=self.P_aMi,
            fracpos_ac=self.fracpos_ac,
            atomdist=self.atomdist,
            kpt_c=self.kpt_c,
            spin=self.spin,
            q=self.q,
            k=self.k,
            weight=self.weight,
            ncomponents=self.ncomponents)

    def move(self,
             fracpos_ac: Array2D,
             atomdist: AtomDistribution) -> None:
        ...

    def force_contribution(self, potential: Potential, F_av: Array2D):
        from gpaw.new.lcao.forces import add_force_contributions
        add_force_contributions(self, potential, F_av)
        return F_av



import pytest
import numpy as np
from ase.build import bulk
from ase.dft.bee import BEEFEnsemble, readbee
from gpaw import GPAW, Mixer, PW
from gpaw.test import gen
from gpaw.mpi import world
import _gpaw


@pytest.mark.mgga
@pytest.mark.libxc
@pytest.mark.slow
def test_beef(in_tmp_dir):
    newlibxc = _gpaw.lxcXCFuncNum('MGGA_X_MBEEF') is not None

    gen('Si', xcname='PBEsol')

    results = {'mBEEF': (5.449, 0.056),
               'BEEF-vdW': (5.484, 0.071),
               'mBEEF-vdW': (5.426, 0.025)}

    for xc in ['mBEEF', 'BEEF-vdW', 'mBEEF-vdW']:
        print(xc)
        if not newlibxc and xc[0] == 'm':
            print('Skipped', xc)
            continue

        if xc == 'mBEEF-vdW':
            # Does not work with libxc-4
            continue

        E = []
        V = []
        for a in np.linspace(5.4, 5.5, 3):
            si = bulk('Si', a=a)
            si.calc = GPAW(txt='Si-' + xc + '.txt',
                           mixer=Mixer(0.8, 7, 50.0),
                           xc=xc,
                           kpts=[2, 2, 2],
                           mode=PW(200))
            E.append(si.get_potential_energy())
            ens = BEEFEnsemble(si.calc, verbose=False)
            ens.get_ensemble_energies(200)
            ens.write('Si-{}-{:.3f}'.format(xc, a))
            V.append(si.get_volume())

        p = np.polyfit(V, E, 2)
        v0 = np.roots(np.polyder(p))[0]
        a = (v0 * 4)**(1 / 3)

        a0, da0 = results[xc]

        assert abs(a - a0) < 0.002, (xc, a, a0)

        if world.rank == 0:
            E = []
            for a in np.linspace(5.4, 5.5, 3):
                e = readbee('Si-{}-{:.3f}'.format(xc, a))
                E.append(e)

            A = []
            for energies in np.array(E).T:
                p = np.polyfit(V, energies, 2)
                assert p[0] > 0, (V, E, p)
                v0 = np.roots(np.polyder(p))[0]
                A.append((v0 * 4)**(1 / 3))

            A = np.array(A)
            a = A.mean()
            da = A.std()

            print('a(ref) = {:.3f} +- {:.3f}'.format(a0, da0))
            print('a      = {:.3f} +- {:.3f}'.format(a, da))
            assert abs(a - a0) < 0.01
            assert abs(da - da0) < 0.01



"""Test BLAS matrix-matrix-multiplication interface."""
import numpy as np
from gpaw.utilities.blas import mmm


def test_linalg_mmm():
    def op(o, m):
        if o == 'N':
            return m
        if o == 'T':
            return m.T
        return m.T.conj()

    def matrix(shape, dtype):
        if dtype == float:
            return np.random.random(shape)
        return np.random.random(shape) + 1j * np.random.random(shape)

    for dtype in [float, complex]:
        a = matrix((2, 3), dtype)
        for opa in 'NTC':
            A = op(opa, a)
            B = matrix((A.shape[1], 4), dtype)
            for opb in 'NTC':
                b = op(opb, B).copy()
                C = np.dot(A, B)
                mmm(1, a, opa, b, opb, -1, C)
                assert abs(C).max() < 1e-14



import sys


def test_stdout():
    class Out:
        def write(self, x):
            sys.__stdout__.write(x)
            raise RuntimeError('not silent')

    out, err = sys.stdout, sys.stderr
    sys.stdout = sys.stderr = Out()

    try:
        from gpaw import GPAW, FermiDirac
        from ase import Atom, Atoms

        a = 5.0
        h = 0.2
        calc = GPAW(h=h, nbands=1, kpts=(1, 1, 1),
                    occupations=FermiDirac(width=1e-9),
                    xc='PBE',
                    txt=None)
        hydrogen = Atoms([Atom('H', (a / 2, a / 2, a / 2), magmom=0)],
                         cell=(a, a, a),
                         calculator=calc)
        hydrogen.get_forces()
    except Exception:
        sys.stdout = out
        sys.stderr = err
        raise

    sys.stdout = out
    sys.stderr = err



import numpy as np
import pytest
from gpaw.gpu import cupy as cp


@pytest.mark.gpu
@pytest.mark.serial
def test_basics():
    a = cp.empty(2)
    b = a.get()
    assert isinstance(b, np.ndarray)
    assert b.dtype == float


@pytest.mark.gpu
@pytest.mark.serial
@pytest.mark.xfail
def test_grr():
    a = cp.empty((2, 2))
    a[:] = 2.5
    b = np.float64(3.0) * a
    assert isinstance(b, cp.ndarray)



import os
import sys
import time

import numpy as np

import ase.db
from ase.units import Rydberg
from ase.utils import opencew
from ase.calculators.calculator import kpts2mp
from ase.io import Trajectory
from ase.calculators.abinit import Abinit
from ase.test.tasks.dcdft import DeltaCodesDFTCollection as Collection

collection = Collection()

if len(sys.argv) == 1:
    names = collection.names
else:
    names = [sys.argv[1]]

c = ase.db.connect('dcdft_abinit_hgh.db')

ecut = 250

kptdensity = 16.0  # this is converged
kptdensity = 6.0  # just for testing
width = 0.01
ecutsm = 0.0
fband = 1.5
tolsym = 1.e-12

linspace = (0.98, 1.02, 5)  # eos numpy's linspace
linspacestr = ''.join([str(t) + 'x' for t in linspace])[:-1]

code = 'abinit' + '-' + '_c' + str(ecut) + '_e' + linspacestr
code = (code + '_k' + str(kptdensity) + '_w' + str(width) + '_s' +
        str(ecutsm) + '_t' + str(tolsym))

for name in names:
    # save all steps in one traj file in addition to the database
    # we should only used the database c.reserve, but here
    # traj file is used as another lock ...
    fd = opencew(name + '_' + code + '.traj')
    if fd is None:
        continue
    traj = Trajectory(name + '_' + code + '.traj', 'w')
    atoms = collection[name]
    if name == 'Mn':  # fails to find the right magnetic state
        atoms.set_initial_magnetic_moments([10., 20., -10., -20.])
    if name == 'Co':  # fails to find the right magnetic state
        atoms.set_initial_magnetic_moments([10., 10.])
    if name == 'Ni':  # fails to find the right magnetic state
        atoms.set_initial_magnetic_moments([10., 10., 10., 10.])
    cell = atoms.get_cell()
    kpts = tuple(kpts2mp(atoms, kptdensity, even=True))
    kwargs = {}
    # loop over EOS linspace
    for n, x in enumerate(np.linspace(linspace[0], linspace[1], linspace[2])):
        id = c.reserve(name=name, ecut=ecut,
                       linspacestr=linspacestr,
                       kptdensity=kptdensity, width=width, ecutsm=ecutsm,
                       fband=fband, tolsym=tolsym,
                       x=x)
        if id is None:
            continue
        # perform EOS step
        atoms.set_cell(cell * x, scale_atoms=True)
        # set calculator
        atoms.calc = Abinit(
            pps='hgh.k',  # uses highest valence hgh.k pps
            label=name + '_' + code + '_' + str(n),
            xc='PBE',
            kpts=kpts,
            ecut=ecut * Rydberg,
            occopt=3,
            tsmear=width,
            ecutsm=ecutsm,
            toldfe=1.0e-6,
            nstep=900,
            pawovlp=-1,  # bypass overlap check
            fband=fband,
            # http://forum.abinit.org/viewtopic.php?f=8&t=35
            chksymbreak=0,
            tolsym=tolsym,
            prtwf=0,
            prtden=0,
        )
        atoms.calc.set(**kwargs)  # remaining calc keywords
        t = time.time()
        atoms.get_potential_energy()
        c.write(atoms,
                name=name, ecut=ecut,
                linspacestr=linspacestr,
                kptdensity=kptdensity, width=width, ecutsm=ecutsm,
                fband=fband, tolsym=tolsym,
                x=x,
                time=time.time() - t)
        traj.write(atoms)
        wfk = name + '_' + code + '_' + str(n) + 'o_WFK'
        if os.path.exists(wfk):
            os.remove(wfk)
        del c[id]



"""Tran-Blaha potential.

From:

    Accurate Band Gaps of Semiconductors and Insulators
    with a Semilocal Exchange-Correlation Potential

    Fabien Tran and Peter Blaha

    PRL 102, 226401 (2009)

    DOI: 10.1103/PhysRevLett.102.226401

"""
import numpy as np
from ase.utils import seterr

from gpaw.xc.mgga import MGGA, weight_n
from gpaw.xc.libxc import LibXC
from gpaw.fd_operators import Laplace


class TB09Kernel:
    name = 'TB09'
    type = 'MGGA'
    alpha = -0.012
    beta = 1.023

    def __init__(self, c=None):
        self.tb09 = LibXC('MGGA_X_TB09').xc.tb09
        self.ldac = LibXC('LDA_C_PW')

        self.fixedc = c is not None  # calculate c or use fixed value
        self.c = c  # amount of "exact exchange"
        self.n = 0  # Lebedev quadrature point number (0-49)
        self.sign = 1.0  # sign of PAW correction: +1 for AE and -1 for PS
        self.I = None  # integral from Eq. (3)

    def calculate(self, e_g, n_sg, dedn_sg, sigma_xg,
                  dedsigma_xg, tau_sg, dedtau_sg):
        ns = len(n_sg)
        n_sg[n_sg < 1e-6] = 1e-6

        if n_sg.ndim == 4:
            if not self.fixedc:
                if self.c is None:
                    # We don't have the integral yet - just use 1.0:
                    self.c = 1.0
                else:
                    self.I = self.world.sum(self.I)
                    self.c = (self.alpha + self.beta *
                              (self.I / self.gd.volume)**0.5)

                # Start calculation of c for use in the next SCF step:
                if ns == 1:
                    gradn_g = sigma_xg[0]**0.5
                else:
                    gradn_g = (sigma_xg[0] +
                               2 * sigma_xg[1] +
                               sigma_xg[2])**0.5
                self.I = self.gd.integrate(gradn_g / n_sg.sum(0))
                # The domain is not distributed like the PAW corrections:
                self.I /= self.world.size

            lapl_sg = self.gd.empty(ns)
            for n_g, lapl_g in zip(n_sg, lapl_sg):
                self.lapl.apply(n_g, lapl_g)

        else:
            rgd = self.rgd
            lapl_sg = []
            for n_Lg in self.n_sLg:
                lapl_g = rgd.laplace(np.dot(self.Y_L, n_Lg))
                l = 0
                L1 = 0
                while L1 < len(self.Y_L):
                    L2 = L1 + 2 * l + 1
                    n_g = np.dot(self.Y_L[L1:L2], n_Lg[L1:L2])
                    with seterr(divide='ignore', invalid='ignore'):
                        lapl_g -= l * (l + 1) * n_g / rgd.r_g**2
                    lapl_g[0] = 0.0
                    L1 = L2
                    l += 1
                lapl_sg.append(lapl_g)

            if not self.fixedc:
                # PAW corrections to integral:
                w = self.sign * weight_n[self.n]
                if ns == 1:
                    gradn_g = sigma_xg[0]**0.5
                else:
                    gradn_g = (sigma_xg[0] +
                               2 * sigma_xg[1] +
                               sigma_xg[2])**0.5
                self.I += w * rgd.integrate(gradn_g / n_sg.sum(0))

                self.n += 1
                if self.n == len(weight_n):
                    self.n = 0
                    self.sign = -self.sign

        # dedn_sg[:] = 0.0
        sigma_xg[sigma_xg < 1e-10] = 1e-10
        tau_sg[tau_sg < 1e-10] = 1e-10

        for n_g, sigma_g, lapl_g, tau_g, v_g in zip(n_sg, sigma_xg[::2],
                                                    lapl_sg, tau_sg, dedn_sg):
            self.tb09(self.c, n_g.ravel(), sigma_g, lapl_g, tau_g, v_g,
                      dedsigma_xg)

        self.ldac.calculate(e_g, n_sg, dedn_sg)
        e_g[:] = 0.0

        dedsigma_xg[:] = 0.0
        dedtau_sg[:] = 0.0


class TB09(MGGA):
    def __init__(self, c=None, stencil=2):
        MGGA.__init__(self, TB09Kernel(c), stencil=stencil)

    def get_setup_name(self):
        return 'LDA'

    def initialize(self, dens, ham, wfs):
        MGGA.initialize(self, dens, ham, wfs)
        self.kernel.world = wfs.world
        self.kernel.gd = dens.finegd
        self.kernel.lapl = Laplace(dens.finegd)

    def create_mgga_radial_calculator(self):
        rcalc = MGGA.create_mgga_radial_calculator(self)

        def f(rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv, n):
            self.kernel.n_sLg = n_sLg
            self.kernel.Y_L = Y_L
            self.kernel.rgd = rgd
            return rcalc(rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv, n)

        return f

    def apply_orbital_dependent_hamiltonian(self, kpt, psit_xG,
                                            Htpsit_xG, dH_asp):
        pass

    @property
    def c(self):
        """Amount of "exact exchange"."""
        return self.kernel.c



import numpy as np
from ase.geometry import cell_to_cellpar
from ase.units import Bohr

from gpaw.fd_operators import Laplace
from gpaw.poisson import _PoissonSolver, create_poisson_solver
from gpaw.transformers import Transformer
from gpaw.utilities.extend_grid import (deextend_array, extend_array,
                                        extended_grid_descriptor)


def ext_gd(gd, **kwargs):
    egd, _, _ = extended_grid_descriptor(gd, **kwargs)
    return egd


class ExtraVacuumPoissonSolver(_PoissonSolver):
    """Wrapper around PoissonSolver extending the vacuum size.

       Poisson equation is solved on the large grid defined by `gpts`
       using `poissonsolver_large`.

       If `coarses` is given, then the large grid is first coarsed
       `coarses` times from to the original grid. Then, the coarse
       potential is used to correct the boundary conditions
       of the potential calculated on the original (small, fine)
       grid by `poissonsolver_small`.

       The parameters `nn_*` control the finite difference stencils
       used in the coarsening, refining, and Laplace.

       If the parameter `use_aux_grid` is `True`, an auxiliary
       medium-sized grid is used between the large and small grids.
       The parameter does not affect the result but can be used to
       achieve speed-up depending on the grid sizes.
       """

    def __init__(self, gpts, poissonsolver_large,
                 poissonsolver_small=None, coarses=0,
                 nn_coarse=3, nn_refine=3, nn_laplace=3,
                 use_aux_grid=True):
        # TODO: Alternative options: vacuum size and h
        self.N_large_fine_c = np.array(gpts, dtype=int)
        self.Ncoar = coarses  # coar == coarse
        if self.Ncoar > 0:
            self.use_coarse = True
        else:
            self.use_coarse = False
        self.ps_large_coar = create_poisson_solver(poissonsolver_large)
        if self.use_coarse:
            self.ps_small_fine = create_poisson_solver(poissonsolver_small)
        else:
            assert poissonsolver_small is None
        self.nn_coarse = nn_coarse
        self.nn_refine = nn_refine
        self.nn_laplace = nn_laplace
        self.use_aux_grid = use_aux_grid
        self._initialized = False

    def set_grid_descriptor(self, gd):
        # If non-periodic boundary conditions is used,
        # there is problems with auxiliary grid.
        # Maybe with use_aux_grid=False it would work?
        if gd.pbc_c.any():
            raise NotImplementedError('Only non-periodic boundary '
                                      'conditions are tested')

        self.gd_small_fine = gd
        assert np.all(self.gd_small_fine.N_c <= self.N_large_fine_c), \
            'extended grid has to be larger than the original one'

        if self.use_coarse:
            # 1.1. Construct coarse chain on the small grid
            self.coarser_i = []
            gd = self.gd_small_fine
            N_c = self.N_large_fine_c.copy()
            for i in range(self.Ncoar):
                gd2 = gd.coarsen()
                self.coarser_i.append(Transformer(gd, gd2, self.nn_coarse))
                N_c //= 2
                gd = gd2
            self.gd_small_coar = gd
        else:
            self.gd_small_coar = self.gd_small_fine
            N_c = self.N_large_fine_c

        # 1.2. Construct coarse extended grid
        self.gd_large_coar = ext_gd(self.gd_small_coar, N_c=N_c)

        # Initialize poissonsolvers
        self.ps_large_coar.set_grid_descriptor(self.gd_large_coar)
        if not self.use_coarse:
            return
        self.ps_small_fine.set_grid_descriptor(self.gd_small_fine)

        if self.use_aux_grid:
            # 2.1. Construct an auxiliary grid that is the small grid plus
            # a buffer region allowing Laplace and refining
            # with the used stencils
            buf = self.nn_refine
            for i in range(self.Ncoar):
                buf = 2 * buf + self.nn_refine
            buf += self.nn_laplace
            div = 2**self.Ncoar
            if buf % div != 0:
                buf += div - buf % div
            N_c = self.gd_small_fine.N_c + 2 * buf
            if np.any(N_c > self.N_large_fine_c):
                self.use_aux_grid = False
                N_c = self.N_large_fine_c
            self.gd_aux_fine = ext_gd(self.gd_small_fine, N_c=N_c)
        else:
            self.gd_aux_fine = ext_gd(self.gd_small_fine,
                                      N_c=self.N_large_fine_c)

        # 2.2 Construct Laplace on the aux grid
        self.laplace_aux_fine = Laplace(self.gd_aux_fine, - 0.25 / np.pi,
                                        self.nn_laplace)

        # 2.3 Construct refine chain
        self.refiner_i = []
        gd = self.gd_aux_fine
        N_c = gd.N_c.copy()
        for i in range(self.Ncoar):
            gd2 = gd.coarsen()
            self.refiner_i.append(Transformer(gd2, gd, self.nn_refine))
            N_c //= 2
            gd = gd2
        self.refiner_i = self.refiner_i[::-1]
        self.gd_aux_coar = gd

        if self.use_aux_grid:
            # 2.4 Construct large coarse grid from aux grid
            self.gd_large_coar_from_aux = ext_gd(self.gd_aux_coar,
                                                 N_c=self.gd_large_coar.N_c)
            # Check the consistency of the grids
            gd1 = self.gd_large_coar
            gd2 = self.gd_large_coar_from_aux
            assert np.all(gd1.N_c == gd2.N_c)
            assert np.allclose(gd1.h_cv, gd2.h_cv, rtol=0, atol=1e-12)

        self._initialized = False

    def _init(self):
        if self._initialized:
            return
        # Allocate arrays
        self.phi_large_coar_g = self.gd_large_coar.zeros()
        self._initialized = True

        # Initialize poissonsolvers
        # self.ps_large_coar._init()
        # if not self.use_coarse:
        #     return
        # self.ps_small_fine._init()

    def solve(self, phi, rho, **kwargs):
        self._init()
        phi_small_fine_g = phi
        rho_small_fine_g = rho.copy()

        if self.use_coarse:
            # 1.1. Coarse rho on the small grid
            tmp_g = rho_small_fine_g
            for coarser in self.coarser_i:
                tmp_g = coarser.apply(tmp_g)
            rho_small_coar_g = tmp_g
        else:
            rho_small_coar_g = rho_small_fine_g

        # 1.2. Extend rho to the large grid
        rho_large_coar_g = self.gd_large_coar.zeros()
        extend_array(self.gd_small_coar, self.gd_large_coar,
                     rho_small_coar_g, rho_large_coar_g)

        # 1.3 Solve potential on the large coarse grid
        niter_large = self.ps_large_coar.solve(self.phi_large_coar_g,
                                               rho_large_coar_g, **kwargs)
        rho_large_coar_g = None

        if not self.use_coarse:
            deextend_array(self.gd_small_fine, self.gd_large_coar,
                           phi_small_fine_g, self.phi_large_coar_g)
            return niter_large

        if self.use_aux_grid:
            # 2.1 De-extend the potential to the auxiliary grid
            phi_aux_coar_g = self.gd_aux_coar.empty()
            deextend_array(self.gd_aux_coar, self.gd_large_coar_from_aux,
                           phi_aux_coar_g, self.phi_large_coar_g)
        else:
            phi_aux_coar_g = self.phi_large_coar_g

        # 3.1 Refine the potential
        tmp_g = phi_aux_coar_g
        for refiner in self.refiner_i:
            tmp_g = refiner.apply(tmp_g)
        phi_aux_coar_g = None
        phi_aux_fine_g = tmp_g

        # 3.2 Calculate the corresponding density with Laplace
        # (the refined coarse density would not accurately match with
        # the potential)
        rho_aux_fine_g = self.gd_aux_fine.empty()
        self.laplace_aux_fine.apply(phi_aux_fine_g, rho_aux_fine_g)

        # 3.3 De-extend the potential and density to the small grid
        cor_phi_small_fine_g = self.gd_small_fine.empty()
        deextend_array(self.gd_small_fine, self.gd_aux_fine,
                       cor_phi_small_fine_g, phi_aux_fine_g)
        phi_aux_fine_g = None
        cor_rho_small_fine_g = self.gd_small_fine.empty()
        deextend_array(self.gd_small_fine, self.gd_aux_fine,
                       cor_rho_small_fine_g, rho_aux_fine_g)
        rho_aux_fine_g = None

        # 3.4 Remove the correcting density and potential
        rho_small_fine_g -= cor_rho_small_fine_g
        phi_small_fine_g -= cor_phi_small_fine_g

        # 3.5 Solve potential on the small grid
        niter_small = self.ps_small_fine.solve(phi_small_fine_g,
                                               rho_small_fine_g, **kwargs)

        # 3.6 Correct potential and density
        phi_small_fine_g += cor_phi_small_fine_g
        # rho_small_fine_g += cor_rho_small_fine_g

        return (niter_large, niter_small)

    def estimate_memory(self, mem):
        self.ps_large_coar.estimate_memory(mem.subnode('Large grid Poisson'))
        if self.use_coarse:
            ps = self.ps_small_fine
            ps.estimate_memory(mem.subnode('Small grid Poisson'))
        mem.subnode('Large coarse phi', self.gd_large_coar.bytecount())
        tmp = max(self.gd_small_fine.bytecount(),
                  self.gd_large_coar.bytecount())
        if self.use_coarse:
            tmp = max(tmp,
                      self.gd_aux_coar.bytecount(),
                      self.gd_aux_fine.bytecount() * 2 +
                      self.gd_small_fine.bytecount(),
                      self.gd_aux_fine.bytecount() +
                      self.gd_small_fine.bytecount() * 2)
        mem.subnode('Temporary arrays', tmp)

    def get_description(self):
        line = '%s with ' % self.__class__.__name__
        if self.use_coarse:
            line += 'large and small grids'
        else:
            line += 'large grid'
        lines = [line]

        def add_line(line, pad=0):
            lines.extend(['%s%s' % (' ' * pad, line)])

        def get_cell(ps):
            descr = ps.get_description().replace('\n', '\n%s' % (' ' * 8))
            add_line('Poisson solver: %s' % descr, 8)
            if hasattr(ps, 'gd'):
                gd = ps.gd
                par = cell_to_cellpar(gd.cell_cv * Bohr)
                h_eff = gd.dv**(1.0 / 3.0) * Bohr
                l1 = '{:8d} x {:8d} x {:8d} points'.format(*gd.N_c)
                l2 = '{:8.2f} x {:8.2f} x {:8.2f} AA'.format(*par[:3])
                l3 = 'Effective grid spacing dv^(1/3) = {:.4f}'.format(h_eff)
                add_line('Grid: %s' % l1, 8)
                add_line('      %s' % l2, 8)
                add_line('      %s' % l3, 8)

        add_line('Large grid:', 4)
        get_cell(self.ps_large_coar)

        if self.use_coarse:
            add_line('Small grid:', 4)
            get_cell(self.ps_small_fine)

        return '\n'.join(lines)

    def todict(self):
        d = {'name': self.__class__.__name__}
        d['gpts'] = self.N_large_fine_c
        d['coarses'] = self.Ncoar
        d['nn_coarse'] = self.nn_coarse
        d['nn_refine'] = self.nn_refine
        d['nn_laplace'] = self.nn_laplace
        d['use_aux_grid'] = self.use_aux_grid
        d['poissonsolver_large'] = self.ps_large_coar.todict()
        if self.use_coarse:
            d['poissonsolver_small'] = self.ps_small_fine.todict()
        else:
            d['poissonsolver_small'] = None
        return d



from __future__ import annotations
import subprocess
import sys
from pathlib import Path

import numpy as np
from ase.io import write
from ase.units import Bohr
from gpaw.new.ase_interface import GPAW


def read_bader_charges(filename: str | Path = 'ACF.dat') -> np.ndarray:
    path = Path(filename)
    charges = []
    with path.open() as fd:
        for line in fd:
            words = line.split()
            if len(words) == 7:
                charges.append(float(words[4]))
    return np.array(charges)


def main(gpw_file_name: str):
    calc = GPAW(gpw_file_name)
    dens = calc.calculation.densities()
    n_sR = dens.all_electron_densities(grid_spacing=0.05)
    # NOTE: Ignoring ASE's hint for **kwargs in write() because it is wrong:
    write('density.cube',
          calc.atoms,
          data=n_sR.data.sum(axis=0) * Bohr**3)  # type: ignore
    subprocess.run('bader -p all_atom density.cube'.split())
    ne = n_sR.integrate().sum()
    print(f'{ne:.6f} electrons')
    charges = calc.atoms.numbers - read_bader_charges()
    for symbol, charge in zip(calc.atoms.symbols, charges):
        print(f'{symbol:2} {charge:10.6f} |e|')


if __name__ == '__main__':
    main(sys.argv[1])




import numpy as np
from gpaw import debug
from gpaw.mpi import world, serial_comm, _Communicator, SerialCommunicator


def test_mpicomm():
    even_comm = world.new_communicator(np.arange(0, world.size, 2))
    if world.size > 1:
        odd_comm = world.new_communicator(np.arange(1, world.size, 2))
    else:
        odd_comm = None

    if world.rank % 2 == 0:
        assert odd_comm is None
        comm = even_comm
    else:
        assert even_comm is None
        comm = odd_comm

    hasmpi = False
    try:
        import _gpaw
        hasmpi = hasattr(_gpaw, 'Communicator') and world.size > 1
    except (ImportError, AttributeError):
        pass

    assert world.parent is None
    assert comm.parent is world
    if hasmpi:
        assert comm.parent.get_c_object() is world.get_c_object()
        assert comm.get_c_object().parent is world.get_c_object()

    commranks = np.arange(world.rank % 2, world.size, 2)
    assert np.all(comm.get_members() == commranks)
    assert comm.get_members()[comm.rank] == world.rank

    subcomm = comm.new_communicator(np.array([comm.rank]))
    assert subcomm.parent is comm
    assert subcomm.rank == 0 and subcomm.size == 1
    assert subcomm.get_members().item() == comm.rank

    if debug:
        assert isinstance(world, _Communicator)
        assert isinstance(comm, _Communicator)
        assert isinstance(subcomm, _Communicator)
    elif world is serial_comm:
        assert isinstance(world, SerialCommunicator)
        assert isinstance(comm, SerialCommunicator)
        assert isinstance(subcomm, SerialCommunicator)
    elif hasmpi:
        assert isinstance(world, _gpaw.Communicator)
        assert isinstance(comm, _gpaw.Communicator)
        assert isinstance(subcomm, _gpaw.Communicator)



from gpaw.response.wgg import choose_parallelization


def test_parallelizations():
    assert choose_parallelization(131, 1455, 160) == (10, 4, 4)
    assert choose_parallelization(470, 10000, 1) == (1, 1, 1)
    assert choose_parallelization(1, 10000, 256) == (1, 16, 16)
    assert choose_parallelization(470, 10000, 256) == (16, 4, 4)



from __future__ import annotations

import numpy as np


class PairTransitions:
    """Bookkeeping object for transitions in band and spin indices.

    All transitions between different band and spin indices (for a given pair
    of k-points k and k + q) are accounted for via single transition index t,

    t (composite transition index): (n, s) -> (n', s')
    """

    def __init__(self, n1_t, n2_t, s1_t, s2_t):
        """Construct the PairTransitions object.

        Parameters
        ----------
        n1_t : np.array
            Band index of k-point k for each transition t.
        n2_t : np.array
            Band index of k-point k + q for each transition t.
        s1_t : np.array
            Spin index of k-point k for each transition t.
        s2_t : np.array
            Spin index of k-point k + q for each transition t.
        """
        self.n1_t = n1_t
        self.n2_t = n2_t
        self.s1_t = s1_t
        self.s2_t = s2_t

        assert len(n2_t) == len(self)
        assert len(s1_t) == len(self)
        assert len(s2_t) == len(self)

    def __len__(self):
        return len(self.n1_t)

    def get_spin_indices(self):
        return self.s1_t, self.s2_t

    def get_band_and_spin_indices(self):
        return self.n1_t, self.n2_t, self.s1_t, self.s2_t

    def get_intraband_mask(self):
        """Get mask for selecting intraband transitions."""
        intraband_t = (self.n1_t == self.n2_t) & (self.s1_t == self.s2_t)
        return intraband_t

    @classmethod
    def from_transitions_domain_arguments(cls, spincomponent,
                                          nbands, nocc1, nocc2, nspins,
                                          bandsummation) -> PairTransitions:
        """Generate the band and spin transitions integration domain.

        The integration domain is determined by the spin rotation (from spin
        index s to spin index s'), the number of bands and spins in the
        underlying ground state calculation as well as the band summation
        scheme.

        The integration domain automatically excludes transitions between two
        occupied bands and two unoccupied bands respectively.

        Parameters
        ----------
        spincomponent : str
            Spin component (μν) of the pair function.
            Currently, '00', 'uu', 'dd', '+-' and '-+' are implemented.
        nbands : int
            Maximum band index to include.
        nocc1 : int
            Number of completely filled bands in the ground state calculation
        nocc2 : int
            Number of non-empty bands in the ground state calculation
        nspins : int
            Number of spin channels in the ground state calculation (1 or 2)
        bandsummation : str
            Band (and spin) summation scheme for pairs of Kohn-Sham orbitals
            'pairwise': sum over pairs of bands (and spins)
            'double': double sum over band (and spin) indices.
        """
        n1_M, n2_M = get_band_transitions_domain(bandsummation, nbands,
                                                 nocc1=nocc1,
                                                 nocc2=nocc2)
        s1_S, s2_S = get_spin_transitions_domain(bandsummation,
                                                 spincomponent, nspins)

        n1_t, n2_t, s1_t, s2_t = transitions_in_composite_index(n1_M, n2_M,
                                                                s1_S, s2_S)

        return cls(n1_t, n2_t, s1_t, s2_t)


def get_band_transitions_domain(bandsummation, nbands, nocc1=None, nocc2=None):
    """Get all pairs of bands to sum over

    Parameters
    ----------
    bandsummation : str
        Band summation method
    nbands : int
        number of bands
    nocc1 : int
        number of completely filled bands
    nocc2 : int
        number of non-empty bands

    Returns
    -------
    n1_M : ndarray
        band index 1, M = (n1, n2) composite index
    n2_M : ndarray
        band index 2, M = (n1, n2) composite index
    """
    _get_band_transitions_domain =\
        create_get_band_transitions_domain(bandsummation)
    n1_M, n2_M = _get_band_transitions_domain(nbands)

    return remove_null_transitions(n1_M, n2_M, nocc1=nocc1, nocc2=nocc2)


def create_get_band_transitions_domain(bandsummation):
    """Creator component deciding how to carry out band summation."""
    if bandsummation == 'pairwise':
        return get_pairwise_band_transitions_domain
    elif bandsummation == 'double':
        return get_double_band_transitions_domain
    raise ValueError(bandsummation)


def get_double_band_transitions_domain(nbands):
    """Make a simple double sum"""
    n_n = np.arange(0, nbands)
    m_m = np.arange(0, nbands)
    n_nm, m_nm = np.meshgrid(n_n, m_m)
    n_M, m_M = n_nm.flatten(), m_nm.flatten()

    return n_M, m_M


def get_pairwise_band_transitions_domain(nbands):
    """Make a sum over all pairs"""
    n_n = range(0, nbands)
    n_M = []
    m_M = []
    for n in n_n:
        m_m = range(n, nbands)
        n_M += [n] * len(m_m)
        m_M += m_m

    return np.array(n_M), np.array(m_M)


def remove_null_transitions(n1_M, n2_M, nocc1=None, nocc2=None):
    """Remove pairs of bands, between which transitions are impossible"""
    n1_newM = []
    n2_newM = []
    for n1, n2 in zip(n1_M, n2_M):
        if nocc1 is not None and (n1 < nocc1 and n2 < nocc1):
            continue  # both bands are fully occupied
        elif nocc2 is not None and (n1 >= nocc2 and n2 >= nocc2):
            continue  # both bands are completely unoccupied
        n1_newM.append(n1)
        n2_newM.append(n2)

    return np.array(n1_newM), np.array(n2_newM)


def get_spin_transitions_domain(bandsummation, spincomponent, nspins):
    """Get structure of the sum over spins

    Parameters
    ----------
    bandsummation : str
        Band summation method
    spincomponent : str
        Spin component (μν) of the pair function.
        Currently, '00', 'uu', 'dd', '+-' and '-+' are implemented.
    nspins : int
        number of spin channels in ground state calculation

    Returns
    -------
    s1_s : ndarray
        spin index 1, S = (s1, s2) composite index
    s2_S : ndarray
        spin index 2, S = (s1, s2) composite index
    """
    _get_spin_transitions_domain =\
        create_get_spin_transitions_domain(bandsummation)
    return _get_spin_transitions_domain(spincomponent, nspins)


def create_get_spin_transitions_domain(bandsummation):
    """Creator component deciding how to carry out spin summation."""
    if bandsummation == 'pairwise':
        return get_pairwise_spin_transitions_domain
    elif bandsummation == 'double':
        return get_double_spin_transitions_domain
    raise ValueError(bandsummation)


def get_double_spin_transitions_domain(spincomponent, nspins):
    """Usual spin rotations forward in time"""
    if nspins == 1:
        if spincomponent == '00':
            s1_S = [0]
            s2_S = [0]
        else:
            raise ValueError(spincomponent, nspins)
    else:
        if spincomponent == '00':
            s1_S = [0, 1]
            s2_S = [0, 1]
        elif spincomponent == 'uu':
            s1_S = [0]
            s2_S = [0]
        elif spincomponent == 'dd':
            s1_S = [1]
            s2_S = [1]
        elif spincomponent == '+-':
            s1_S = [0]  # spin up
            s2_S = [1]  # spin down
        elif spincomponent == '-+':
            s1_S = [1]  # spin down
            s2_S = [0]  # spin up
        else:
            raise ValueError(spincomponent)

    return np.array(s1_S), np.array(s2_S)


def get_pairwise_spin_transitions_domain(spincomponent, nspins):
    """In a sum over pairs, transitions including a spin rotation may have to
    include terms, propagating backwards in time."""
    if spincomponent in ['+-', '-+']:
        assert nspins == 2
        return np.array([0, 1]), np.array([1, 0])
    else:
        return get_double_spin_transitions_domain(spincomponent, nspins)


def transitions_in_composite_index(n1_M, n2_M, s1_S, s2_S):
    """Use a composite index t for transitions (n, s) -> (n', s')."""
    n1_MS, s1_MS = np.meshgrid(n1_M, s1_S)
    n2_MS, s2_MS = np.meshgrid(n2_M, s2_S)
    return n1_MS.flatten(), n2_MS.flatten(), s1_MS.flatten(), s2_MS.flatten()



from gpaw.mpi import world
from gpaw.blacs import BlacsGrid, Redistributor


def test_parallel_submatrix_redist(scalapack):
    if world.size > 1:
        shape = (2, world.size // 2)
    else:
        shape = (1, 1)

    grid = BlacsGrid(world, *shape)

    desc = grid.new_descriptor(12, 8, 2, 3)

    a = desc.zeros()
    a[:] = world.rank

    subdesc = grid.new_descriptor(7, 7, 2, 2)
    b = subdesc.zeros()

    r = Redistributor(grid.comm, desc, subdesc)

    ia = 3
    ja = 2
    ib = 1
    jb = 1
    M = 4
    N = 5

    r.redistribute(a, b, M, N, ia, ja, ib, jb)

    a0 = desc.collect_on_master(a)
    b0 = subdesc.collect_on_master(b)
    if world.rank == 0:
        print(a0)
        print(b0)
        xa = a0[ia:ia + M, ja:ja + N]
        xb = b0[ib:ib + M, jb:jb + N]
        assert (xa == xb).all()



from functools import partial

from ase.utils.timing import timer
import numpy as np

from gpaw import debug
from gpaw.eigensolvers.eigensolver import Eigensolver
from gpaw.matrix import matrix_matrix_multiply as mmm
from gpaw.hybrids import HybridXC
from gpaw.eigensolvers.diagonalizerbackend import (
    ScipyDiagonalizer,
    ScalapackDiagonalizer)


class DummyArray:
    def __getitem__(self, x):
        return np.empty((0, 0))


class Davidson(Eigensolver):
    """Simple Davidson eigensolver

    It is expected that the trial wave functions are orthonormal
    and the integrals of projector functions and wave functions
    ``nucleus.P_uni`` are already calculated.

    Solution steps are:

    * Subspace diagonalization
    * Calculate all residuals
    * Add preconditioned residuals to the subspace and diagonalize
    """

    def __init__(
            self, niter=2):
        Eigensolver.__init__(self)
        self.niter = niter
        self.diagonalizer_backend = None

        self.orthonormalization_required = False
        self.H_NN = DummyArray()
        self.S_NN = DummyArray()
        self.eps_N = DummyArray()

    def __repr__(self):
        return 'Davidson(niter=%d)' % (
            self.niter)

    def todict(self):
        return {'name': 'dav', 'niter': self.niter}

    def initialize(self, wfs):
        Eigensolver.initialize(self, wfs)
        slcomm, nrows, ncols, slsize = wfs.scalapack_parameters

        if wfs.gd.comm.rank == 0 and wfs.bd.comm.rank == 0:
            # Allocate arrays
            B = self.nbands
            self.H_NN = np.zeros((2 * B, 2 * B), self.dtype)
            self.S_NN = np.zeros((2 * B, 2 * B), self.dtype)
            self.eps_N = np.zeros(2 * B)

        if slsize is not None:
            self.diagonalizer_backend = ScalapackDiagonalizer(
                arraysize=self.nbands * 2,
                grid_nrows=nrows,
                grid_ncols=ncols,
                scalapack_communicator=slcomm,
                dtype=self.dtype,
                blocksize=slsize)
        else:
            self.diagonalizer_backend = ScipyDiagonalizer()

    def estimate_memory(self, mem, wfs):
        Eigensolver.estimate_memory(self, mem, wfs)
        nbands = wfs.bd.nbands
        mem.subnode('H_nn', nbands * nbands * mem.itemsize[wfs.dtype])
        mem.subnode('S_nn', nbands * nbands * mem.itemsize[wfs.dtype])
        mem.subnode('H_2n2n', 4 * nbands * nbands * mem.itemsize[wfs.dtype])
        mem.subnode('S_2n2n', 4 * nbands * nbands * mem.itemsize[wfs.dtype])
        mem.subnode('eps_2n', 2 * nbands * mem.floatsize)

    @timer('Davidson')
    def iterate_one_k_point(self, ham, wfs, kpt, weights):
        """Do Davidson iterations for the kpoint"""
        if isinstance(ham.xc, HybridXC):
            self.niter = 1

        bd = wfs.bd
        B = bd.nbands

        H_NN = self.H_NN
        S_NN = self.S_NN
        eps_N = self.eps_N

        def integrate(a_G):
            if wfs.collinear:
                return np.real(wfs.integrate(a_G, a_G, global_integral=False))
            return sum(
                np.real(wfs.integrate(b_G, b_G, global_integral=False))
                for b_G in a_G)

        self.subspace_diagonalize(ham, wfs, kpt)

        psit = kpt.psit
        psit2 = psit.new(buf=wfs.work_array)
        P = kpt.projections
        P2 = P.new()
        P3 = P.new()
        M = wfs.work_matrix_nn
        dS = wfs.setups.dS
        comm = wfs.gd.comm

        is_gridband_master = (comm.rank == 0) and (bd.comm.rank == 0)

        if bd.comm.size > 1:
            M0 = M.new(dist=(bd.comm, 1, 1))
        else:
            M0 = M

        if comm.rank == 0:
            e_N = bd.collect(kpt.eps_n)
            if e_N is not None:
                eps_N[:B] = e_N

        Ht = partial(wfs.apply_pseudo_hamiltonian, kpt, ham)

        if self.keep_htpsit:
            R = psit.new(buf=self.Htpsit_nG)
        else:
            R = psit.apply(Ht)

        self.calculate_residuals(kpt, wfs, ham, psit, P, kpt.eps_n, R, P2)

        precond = self.preconditioner

        for nit in range(self.niter):
            if nit == self.niter - 1:
                error = np.dot(weights, [integrate(R_G) for R_G in R.array])

            for psit_G, R_G, psit2_G in zip(psit.array, R.array, psit2.array):
                ekin = precond.calculate_kinetic_energy(psit_G, kpt)
                precond(R_G, kpt, ekin, out=psit2_G)

            # Calculate projections
            psit2.matrix_elements(wfs.pt, out=P2)

            def copy(M, C_nn):
                comm.sum(M.array, 0)
                if comm.rank == 0:
                    M.complex_conjugate()
                    M.redist(M0)
                    if bd.comm.rank == 0:
                        C_nn[:] = M0.array

            with self.timer('calc. matrices'):
                # <psi2 | H | psi2>
                psit2.matrix_elements(
                    operator=Ht, result=R, out=M, symmetric=True, cc=True)
                ham.dH(P2, out=P3)
                mmm(1.0, P2, 'N', P3, 'C', 1.0, M)  # , symmetric=True)
                copy(M, H_NN[B:, B:])

                # <psi2 | H | psi>
                R.matrix_elements(psit, out=M, cc=True)
                mmm(1.0, P3, 'N', P, 'C', 1.0, M)
                copy(M, H_NN[B:, :B])

                # <psi2 | S | psi2>
                psit2.matrix_elements(out=M, symmetric=True, cc=True)
                dS.apply(P2, out=P3)
                mmm(1.0, P2, 'N', P3, 'C', 1.0, M)
                copy(M, S_NN[B:, B:])

                # <psi2 | S | psi>
                psit2.matrix_elements(psit, out=M, cc=True)
                mmm(1.0, P3, 'N', P, 'C', 1.0, M)
                copy(M, S_NN[B:, :B])

            if is_gridband_master:
                H_NN[:B, :B] = np.diag(eps_N[:B])
                S_NN[:B, :B] = np.eye(B)

            with self.timer('diagonalize'):
                if is_gridband_master and debug:
                    H_NN[np.triu_indices(2 * B, 1)] = 42.0
                    S_NN[np.triu_indices(2 * B, 1)] = 42.0

                try:
                    self.diagonalizer_backend.diagonalize(
                        H_NN, S_NN, eps_N, is_master=is_gridband_master,
                        debug=debug)
                except np.linalg.LinAlgError as ex:
                    raise ValueError(
                        'Too few plane waves or grid points') from ex
            if comm.rank == 0:
                bd.distribute(eps_N[:B], kpt.eps_n)
            comm.broadcast(kpt.eps_n, 0)

            with self.timer('rotate_psi'):
                if comm.rank == 0:
                    if bd.comm.rank == 0:
                        M0.array[:] = H_NN[:B, :B].T
                    M0.redist(M)
                comm.broadcast(M.array, 0)
                mmm(1.0, M, 'N', psit, 'N', 0.0, R)
                mmm(1.0, M, 'N', P, 'N', 0.0, P3)
                if comm.rank == 0:
                    if bd.comm.rank == 0:
                        M0.array[:] = H_NN[B:, :B].T
                    M0.redist(M)
                comm.broadcast(M.array, 0)
                mmm(1.0, M, 'N', psit2, 'N', 1.0, R)
                mmm(1.0, M, 'N', P2, 'N', 1.0, P3)
                psit[:] = R
                P, P3 = P3, P
                kpt.projections = P

            if nit < self.niter - 1:
                psit.apply(Ht, out=R)
                self.calculate_residuals(
                    kpt, wfs, ham, psit, P, kpt.eps_n, R, P2)

        error = wfs.gd.comm.sum(error)
        return error



from math import cos, pi, sin

import numpy as np
import pytest
from ase import Atom, Atoms

import gpaw.mpi as mpi
from gpaw import GPAW
from gpaw.atom.generator2 import generate
from gpaw.poisson import FDPoissonSolver
from gpaw.test import equal
from gpaw.xas import XAS


@pytest.mark.later
def test_corehole_h2o(in_tmp_dir, add_cwd_to_setup_paths):
    # Generate setup for oxygen with half a core-hole:
    gen = generate('O', 8, '2s,s,2p,p,d', [1.2], 1.0, None, 2,
                   core_hole='1s,0.5')
    setup = gen.make_paw_setup('hch1s')
    setup.write_xml()

    a = 5.0
    d = 0.9575
    t = pi / 180 * 104.51
    H2O = Atoms([Atom('O', (0, 0, 0)),
                 Atom('H', (d, 0, 0)),
                 Atom('H', (d * cos(t), d * sin(t), 0))],
                cell=(a, a, a), pbc=False)
    H2O.center()
    calc = GPAW(nbands=10, h=0.2, setups={'O': 'hch1s'},
                experimental={'niter_fixdensity': 2},
                poissonsolver=FDPoissonSolver(use_charge_center=True))
    H2O.calc = calc
    _ = H2O.get_potential_energy()

    if mpi.size == 1:
        xas = XAS(calc)
        x, y = xas.get_spectra()
        e1_n = xas.eps_n
        de1 = e1_n[1] - e1_n[0]

    calc.write('h2o-xas.gpw')

    if mpi.size == 1:
        calc = GPAW('h2o-xas.gpw')
        # poissonsolver=FDPoissonSolver(use_charge_center=True))
        # calc.initialize()
        xas = XAS(calc)
        x, y = xas.get_spectra()
        e2_n = xas.eps_n
        w_n = np.sum(xas.sigma_cn.real**2, axis=0)
        de2 = e2_n[1] - e2_n[0]

        equal(de2, 2.064, 0.005)
        equal(w_n[1] / w_n[0], 2.22, 0.01)

        assert de1 == de2

    if 0:
        import matplotlib.pyplot as plt
        plt.plot(x, y[0])
        plt.plot(x, sum(y))
        plt.show()



"""Module for electron-phonon supercell properties."""

import numpy as np
from typing import Tuple

from ase import Atoms
from ase.parallel import parprint
import ase.units as units
from ase.utils.filecache import MultiFileJSONCache

from gpaw.calculator import GPAW
from gpaw.lcao.tightbinding import TightBinding
from gpaw.utilities import unpack2
from gpaw.utilities.tools import tri2full
from gpaw.typing import ArrayND

from .filter import fourier_filter

sc_version = 1
# v1: saves natom, supercell, g_sNNMM.shape and dtype


class VersionError(Exception):
    """Error raised for wrong cache versions."""
    pass


class Supercell:
    """Class for supercell-related stuff."""

    def __init__(self, atoms: Atoms, supercell_name: str = "supercell",
                 supercell: tuple = (1, 1, 1)) -> None:
        """Initialize supercell class.

        Parameters
        ----------
        atoms: Atoms
            The atoms to work on. Primitive cell.
        supercell_name: str
            User specified name of the generated JSON cache.
            Default is 'supercell'.
        supercell: tuple
            Size of supercell given by the number of repetitions (l, m, n) of
            the small unit cell in each direction.
        """
        self.atoms = atoms
        self.supercell_name = supercell_name
        self.supercell = supercell

    def _calculate_supercell_entry(self, a, v, V1t_sG, dH1_asp, wfs,
                                   dH_asp) -> ArrayND:
        kpt_u = wfs.kpt_u
        setups = wfs.setups
        nao = setups.nao
        bfs = wfs.basis_functions
        dtype = wfs.dtype
        nspins = wfs.nspins
        indices = np.arange(len(self.atoms))

        # Array for different k-point components
        g_sqMM = np.zeros((nspins, len(kpt_u) // nspins, nao, nao), dtype)

        # 1) Gradient of effective potential
        for kpt in kpt_u:
            # Matrix elements
            # Note: somehow this part does not work with gd-parallelisation
            geff_MM = np.zeros((nao, nao), dtype)
            bfs.calculate_potential_matrix(V1t_sG[kpt.s], geff_MM, q=kpt.q)
            tri2full(geff_MM, "L")
            # wfs.gd.comm.sum(geff_MM)
            # print(world.rank, a, v, kpt.k, geff_MM)
            g_sqMM[kpt.s, kpt.q] += geff_MM
            # print(wfs.kd.comm.rank, wfs.gd.comm.rank, wfs.bd.comm.rank,
            #       "\n", geff_MM)

        # 2) Gradient of non-local part (projectors)
        P_aqMi = wfs.P_aqMi
        # 2a) dH^a part has contributions from all other atoms
        for kpt in kpt_u:
            # Matrix elements
            gp_MM = np.zeros((nao, nao), dtype)
            for a_, dH1_sp in dH1_asp.items():
                if a_ not in bfs.my_atom_indices:
                    continue
                dH1_ii = unpack2(dH1_sp[kpt.s])
                P_Mi = P_aqMi[a_][kpt.q]
                gp_MM += P_Mi.conj() @ dH1_ii @ P_Mi.T
            # wfs.gd.comm.sum(gp_MM)
            g_sqMM[kpt.s, kpt.q] += gp_MM

        # 2b) dP^a part has only contributions from the same atoms
        # For the contribution from the derivative of the projectors
        dPdR_aqvMi = wfs.manytci.P_aqMi(indices, derivative=True)
        dH_ii = unpack2(dH_asp[a][kpt.s])
        for kpt in kpt_u:
            gp_MM = np.zeros((nao, nao), dtype)
            if a in bfs.my_atom_indices:
                P_Mi = P_aqMi[a][kpt.q]
                dP_Mi = dPdR_aqvMi[a][kpt.q][v]
                P1HP_MM = dP_Mi.conj() @ dH_ii @ P_Mi.T
                # Matrix elements
                gp_MM += P1HP_MM + P1HP_MM.T.conjugate()
            # wfs.gd.comm.sum(gp_MM)
            # print(world.rank, a,v, kpt.k, bfs.my_atom_indices, gp_MM)
            g_sqMM[kpt.s, kpt.q] += gp_MM

        return g_sqMM

    def calculate_supercell_matrix(
        self, calc: GPAW, fd_name: str = "elph", filter: str = None
    ) -> None:
        """Calculate matrix elements of the el-ph coupling in the LCAO basis.

        This function calculates the matrix elements between LCAOs and local
        atomic gradients of the effective potential. The matrix elements are
        calculated for the supercell used to obtain finite-difference
        approximations to the derivatives of the effective potential wrt to
        atomic displacements.

        The resulting g_xsNNMM is saved into a JSON cache.

        Parameters
        ----------
        calc: GPAW
            LCAO calculator for the calculation of the supercell matrix.
        fd_name: str
            User specified name of the finite difference JSON cache.
            Default is 'elph'.
        filter: str
            Fourier filter atomic gradients of the effective potential. The
            specified components (``normal`` or ``umklapp``) are removed
            (default: None).
        """

        assert calc.wfs.mode == "lcao", "LCAO mode required."
        assert not calc.symmetry.point_group, \
            "Point group symmetry not supported"

        # JSON cache
        supercell_cache = MultiFileJSONCache(self.supercell_name)

        # Supercell atoms
        atoms_N = self.atoms * self.supercell

        # Initialize calculator if required and extract useful quantities
        if (not hasattr(calc.wfs, "S_qMM") or
            not hasattr(calc.wfs.basis_functions, "M_a")):
            calc.initialize(atoms_N)
            calc.initialize_positions(atoms_N)

        # Extract useful objects from the calculator
        wfs = calc.wfs
        gd = calc.wfs.gd
        kd = calc.wfs.kd
        bd = calc.wfs.bd
        nao = wfs.setups.nao
        nspins = wfs.nspins
        # FIXME: Domain parallelisation broken
        assert gd.comm.size == 1
        # FIXME: Band parallelisation broken - M is band parallel
        assert bd.comm.size == 1

        # Calculate finite-difference gradients (in Hartree / Bohr)
        V1t_xsG, dH1_xasp = self.calculate_gradient(fd_name)

        # Equilibrium atomic Hamiltonian matrix (projector coefficients)
        fd_cache = MultiFileJSONCache(fd_name)
        dH_asp = fd_cache["eq"]["dH_all_asp"]

        # Check that the grid is the same as in the calculator
        assert np.all(V1t_xsG.shape[-3:] == (gd.N_c + gd.pbc_c - 1)), \
            "Mismatch in grids."

        # Save basis information, after we checked the data is kosher
        with supercell_cache.lock("basis") as handle:
            if handle is not None:
                basis_info = self.set_basis_info(calc)
                handle.save(basis_info)

        # Fourier filter the atomic gradients of the effective potential
        if filter is not None:
            for s in range(nspins):
                fourier_filter(self.atoms, self.supercell, V1t_xsG[:, s],
                               components=filter)

        if kd.gamma:
            print("WARNING: Gamma-point calculation. \
                   Overlap with neighboring cell cannot be removed")
        else:
            # Bloch to real-space converter
            tb = TightBinding(atoms_N, calc)

        # Calculate < i k | grad H | j k >, i.e. matrix elements in LCAO basis

        # Do each cartesian component separately
        for i, a in enumerate(np.arange(len(self.atoms))):
            for v in range(3):
                # Corresponding array index
                x = 3 * i + v

                # If exist already, don't recompute
                with supercell_cache.lock(str(x)) as handle:
                    if handle is None:
                        continue

                    parprint("%s-gradient of atom %u" %
                             (["x", "y", "z"][v], a))

                    g_sqMM = self._calculate_supercell_entry(
                        a, v, V1t_xsG[x], dH1_xasp[x], wfs, dH_asp
                    )

                    # Extract R_c=(0, 0, 0) block by Fourier transforming
                    if kd.gamma or kd.N_c is None:
                        g_sMM = g_sqMM[:, 0]
                    else:
                        # Convert to array
                        g_sMM_tmp = []
                        for s in range(nspins):
                            g_MM = tb.bloch_to_real_space(g_sqMM[s],
                                                          R_c=(0, 0, 0))
                            g_sMM_tmp.append(g_MM[0])  # [0] because of above
                        g_sMM = np.array(g_sMM_tmp)
                        del g_sMM_tmp

                    # Reshape to global unit cell indices
                    N = np.prod(self.supercell)
                    # Number of basis function in the primitive cell
                    assert (nao % N) == 0, "Alarm ...!"
                    nao_cell = nao // N
                    g_sNMNM = g_sMM.reshape((nspins, N, nao_cell, N, nao_cell))
                    g_sNNMM = g_sNMNM.swapaxes(2, 3).copy()
                    handle.save(g_sNNMM)
                if x == 0:
                    with supercell_cache.lock("info") as handle:
                        if handle is not None:
                            info = {
                                "sc_version": sc_version,
                                "natom": len(self.atoms),
                                "supercell": self.supercell,
                                "gshape": g_sNNMM.shape,
                                "gtype": g_sNNMM.dtype.name,
                            }
                            handle.save(info)

    def set_basis_info(self, *args) -> dict:
        """Store LCAO basis info for atoms in reference cell in attribute.

        Parameters
        ----------
        args: tuple
            If the LCAO calculator is not available (e.g. if the supercell is
            loaded from file), the ``load_supercell_matrix`` member function
            provides the required info as arguments.

        """
        assert len(args) in (1, 2)
        if len(args) == 1:
            calc = args[0]
            setups = calc.wfs.setups
            bfs = calc.wfs.basis_functions
            nao_a = [setups[a].nao for a in range(len(self.atoms))]
            M_a = [bfs.M_a[a] for a in range(len(self.atoms))]
        else:
            M_a = args[0]
            nao_a = args[1]
        return {"M_a": M_a, "nao_a": nao_a}

    @classmethod
    def calculate_gradient(cls, fd_name: str) -> Tuple[ArrayND, list]:
        """Calculate gradient of effective potential and projector coefs.

        This function loads the generated json files and calculates
        finite-difference derivatives.

        Parameters
        ----------
        fd_name: str
            name of finite difference JSON cache
        """
        cache = MultiFileJSONCache(fd_name)
        if "dr_version" not in cache["info"]:
            print("Cache created with old version. Use electronphonon.py")
            raise VersionError
        natom = cache["info"]["natom"]
        delta = cache["info"]["delta"]

        # Array and dict for finite difference derivatives
        V1t_xsG = []
        dH1_xasp = []

        x = 0
        for a in range(natom):
            for v in "xyz":
                name = "%d%s" % (a, v)
                # Potential and atomic density matrix for atomic displacement
                Vtm_sG = cache[name + "-"]["Vt_sG"]
                dHm_asp = cache[name + "-"]["dH_all_asp"]
                Vtp_sG = cache[name + "+"]["Vt_sG"]
                dHp_asp = cache[name + "+"]["dH_all_asp"]

                # FD derivatives in Hartree / Bohr
                V1t_sG = (Vtp_sG - Vtm_sG) / (2 * delta / units.Bohr)
                V1t_xsG.append(V1t_sG)

                dH1_asp = {}
                for atom in dHm_asp.keys():
                    dH1_asp[atom] = (dHp_asp[atom] - dHm_asp[atom]) / (
                        2 * delta / units.Bohr
                    )
                dH1_xasp.append(dH1_asp)
                x += 1
        return np.array(V1t_xsG), dH1_xasp

    @classmethod
    def load_supercell_matrix(cls, name: str = "supercell"
                              ) -> Tuple[ArrayND, dict]:
        """Load supercell matrix from cache.

        Parameters
        ----------
        name: str
            User specified name of the cache.
        """
        supercell_cache = MultiFileJSONCache(name)
        if "sc_version" not in supercell_cache["info"]:
            print("Cache created with old version. Use electronphonon.py")
            raise VersionError
        shape = supercell_cache["info"]["gshape"]
        dtype = supercell_cache["info"]["gtype"]
        natom = supercell_cache["info"]["natom"]
        nx = natom * 3
        g_xsNNMM = np.empty([nx, ] + list(shape), dtype=dtype)
        for x in range(nx):
            g_xsNNMM[x] = supercell_cache[str(x)]
        basis_info = supercell_cache["basis"]
        return g_xsNNMM, basis_info



"""Test the site kernel calculation functionality of the response code"""

# General modules
import pytest
import numpy as np
import scipy.special as sc

# Script modules
from ase.build import bulk

from gpaw import GPAW, PW
from gpaw.response.site_kernels import (SphericalSiteKernels,
                                        CylindricalSiteKernels,
                                        ParallelepipedicSiteKernels,
                                        sinc,
                                        spherical_geometry_factor,
                                        cylindrical_geometry_factor,
                                        parallelepipedic_geometry_factor)
from gpaw.response.pair_functions import get_pw_coordinates


# ---------- Actual tests ---------- #


pytestmark = pytest.mark.kspair


@pytest.mark.ci
def test_spherical_kernel():
    """Check the numerics of the spherical kernel"""
    # ---------- Inputs ---------- #

    # Relative wave vector lengths to check (relative to 1/rc)
    Qrel_Q = np.array([0.,
                       np.pi / 2,
                       np.pi])

    # Expected results (assuming |Q| * rc = 1.)
    Vsphere = 4. * np.pi / 3.
    test_K_Q = Vsphere * np.array([1.,
                                   3 / (np.pi / 2)**3.,
                                   3 / (np.pi)**2.])

    # Spherical radii to check
    nr = 5
    rc_r = np.random.rand(nr)

    # Wave vector directions to check
    nd = 41
    Q_dv = 2. * np.random.rand(nd, 3) - 1.
    Q_dv /= np.linalg.norm(Q_dv, axis=1)[:, np.newaxis]  # normalize

    # ---------- Script ---------- #

    # Set up wave vectors
    Q_Qdv = Qrel_Q[:, np.newaxis, np.newaxis] * Q_dv[np.newaxis, ...]

    for rc in rc_r:
        # Calculate site centered geometry factor with rescaled wave vector
        K_Qd = spherical_geometry_factor(Q_Qdv / rc, rc)
        # Check against expected result
        assert np.allclose(K_Qd, rc**3. * test_K_Q[:, np.newaxis])


@pytest.mark.ci
def test_cylindrical_kernel():
    """Check the numerics of the spherical kernel"""
    # ---------- Inputs ---------- #

    # Relative wave vectors (relative to 1/rc) in radial direction
    Qrhorel_Q1 = np.array([0.] + list(sc.jn_zeros(1, 4)))  # Roots of J1(x)

    # Relative wave vectors (relative to 2/hc) in cylindrical direction
    Qzrel_Q2 = list(np.pi * np.arange(5))  # Roots of sin(x)
    Qzrel_Q2 += list(np.pi * np.arange(4) + np.pi / 2)  # Extrema of sin(x)
    Qzrel_Q2 = np.array(Qzrel_Q2)

    # Expected results for roots of J1 (assuming rc=1. and hc=2.)
    Vcylinder = 2. * np.pi
    nQ2 = 13  # Choose random Q_z r_z
    test_Krho_Q1 = np.array([1., 0., 0., 0., 0.])
    Qzrand_Q2 = 10. * np.random.rand(nQ2)
    sinc_zrand_Q2 = np.sin(Qzrand_Q2) / Qzrand_Q2
    test_Krho_Q1Q2 = Vcylinder * test_Krho_Q1[:, np.newaxis]\
        * sinc_zrand_Q2[np.newaxis, :]

    # Expected results for roots and extrema of sin (assuming rc=1. and hc=2.)
    nQ1 = 15  # Choose random Q_ρ h_c
    test_Kz_Q2 = [1., 0., 0., 0., 0.]  # Nodes in sinc(Q_z h_c)
    test_Kz_Q2 += list(np.array([1., -1., 1., -1.]) / Qzrel_Q2[5:])  # Extrema
    test_Kz_Q2 = np.array(test_Kz_Q2)
    Qrhorand_Q1 = 10. * np.random.rand(nQ1)
    J1term_rhorand_Q1 = 2. * sc.jv(1, Qrhorand_Q1) / Qrhorand_Q1
    test_Kz_Q1Q2 = Vcylinder * J1term_rhorand_Q1[:, np.newaxis]\
        * test_Kz_Q2[np.newaxis, :]

    # Cylinder radii to check
    nr = 5
    rc_r = 3. * np.random.rand(nr)

    # Cylinder height to check
    nh = 3
    hc_h = 4. * np.random.rand(nh)

    # Cylindrical axes to check
    nc = 7
    ez_cv = 2. * np.random.rand(nc, 3) - 1.
    ez_cv /= np.linalg.norm(ez_cv, axis=1)[:, np.newaxis]

    # Wave vector directions in-plane to check. Generated through the cross
    # product of a random direction with the cylindrical axis
    nd = 11
    Qrho_dv = 2. * np.random.rand(nd, 3) - 1.
    Qrho_cdv = np.cross(Qrho_dv[np.newaxis, ...], ez_cv[:, np.newaxis, :])
    Qrho_cdv /= np.linalg.norm(Qrho_cdv, axis=-1)[..., np.newaxis]  # normalize

    # ---------- Script ---------- #

    for rc in rc_r:
        for hc in hc_h:
            # Set up wave vectors for radial tests
            Qrho_cdQ1v = Qrhorel_Q1[np.newaxis, np.newaxis, :, np.newaxis]\
                * Qrho_cdv[..., np.newaxis, :] / rc
            Qrho_cQ2v = Qzrand_Q2[np.newaxis, :, np.newaxis]\
                * ez_cv[:, np.newaxis, :] / (hc / 2.)
            Qrho_cdQ1Q2v = Qrho_cdQ1v[..., np.newaxis, :]\
                + Qrho_cQ2v[:, np.newaxis, np.newaxis, ...]

            # Set up wave vectors for cylindrical tests
            Qz_cdQ1v = Qrhorand_Q1[np.newaxis, np.newaxis, :, np.newaxis]\
                * Qrho_cdv[..., np.newaxis, :] / rc
            Qz_cQ2v = Qzrel_Q2[np.newaxis, :, np.newaxis]\
                * ez_cv[:, np.newaxis, :] / (hc / 2.)
            Qz_cdQ1Q2v = Qz_cdQ1v[..., np.newaxis, :]\
                + Qz_cQ2v[:, np.newaxis, np.newaxis, ...]

            # Test one cylindrical direction at a time
            for ez_v, Qrho_dQ1Q2v, Qz_dQ1Q2v in zip(ez_cv,
                                                    Qrho_cdQ1Q2v, Qz_cdQ1Q2v):
                # Calculate geometry factors
                Krho_dQ1Q2 = cylindrical_geometry_factor(Qrho_dQ1Q2v,
                                                         ez_v, rc, hc)
                Kz_dQ1Q2 = cylindrical_geometry_factor(Qz_dQ1Q2v,
                                                       ez_v, rc, hc)

                # Check against expected result
                assert np.allclose(Krho_dQ1Q2, rc**2. * hc / 2.
                                   * test_Krho_Q1Q2[np.newaxis, ...],
                                   atol=1.e-8)
                assert np.allclose(Kz_dQ1Q2, rc**2. * hc / 2.
                                   * test_Kz_Q1Q2[np.newaxis, ...],
                                   atol=1.e-8)


@pytest.mark.ci
def test_parallelepipedic_kernel():
    """Check the numerics of the parallelepipedic site kernel."""
    # ---------- Inputs ---------- #

    # Relative wave vectors to check and corresponding sinc(x/2)
    Qrel_Q = np.pi * np.arange(5)
    sinchalf_Q = np.array([1., 2. / np.pi, 0., - 2. / (3. * np.pi), 0.])

    # Random parallelepipedic cell vectors to check
    nC = 9
    cell_Ccv = 2. * np.random.rand(nC, 3, 3) - 1.
    volume_C = np.abs(np.linalg.det(cell_Ccv))
    # Normalize the cell volume
    cell_Ccv /= (volume_C**(1 / 3))[:, np.newaxis, np.newaxis]

    # Transverse wave vector components to check. Generated through the cross
    # product of a random direction with the first cell axis.
    v0_Cv = cell_Ccv[:, 0, :].copy()
    v0_C = np.linalg.norm(v0_Cv, axis=-1)  # Length of primary vector
    v0n_Cv = v0_Cv / v0_C[:, np.newaxis]  # Normalize
    nd = 11
    Q_dv = 2. * np.random.rand(nd, 3) - 1.
    Q_dv[0, :] = np.array([0., 0., 0.])  # Check also parallel Q-vector
    Q_Cdv = np.cross(Q_dv[np.newaxis, ...], v0_Cv[:, np.newaxis, :])

    # Volumes to test
    nV = 7
    Vparlp_V = 10. * np.random.rand(nV)

    # ---------- Script ---------- #

    # Rescale cell
    cell_CVcv = cell_Ccv[:, np.newaxis, ...]\
        * (Vparlp_V**(1 / 3.))[np.newaxis, :, np.newaxis, np.newaxis]

    # Rescale primary vector to let Q.a follow Qrel
    Qrel_CQ = Qrel_Q[np.newaxis, :] / v0_C[:, np.newaxis]
    Qrel_CVQ = Qrel_CQ[:, np.newaxis, :]\
        / (Vparlp_V**(1 / 3.))[np.newaxis, :, np.newaxis]
    # Generate Q-vectors
    Q_CVdQv = Qrel_CVQ[..., np.newaxis, :, np.newaxis]\
        * v0n_Cv[:, np.newaxis, np.newaxis, np.newaxis, :]\
        + Q_Cdv[:, np.newaxis, :, np.newaxis, :]

    # Generate test values
    sinchalf_CVdQ = sinc(np.sum(cell_CVcv[..., np.newaxis, np.newaxis, 1, :]
                                * Q_CVdQv, axis=-1) / 2)\
        * sinc(np.sum(cell_CVcv[..., np.newaxis, np.newaxis, 2, :]
                      * Q_CVdQv, axis=-1) / 2)
    test_Theta_CVdQ = Vparlp_V[np.newaxis, :, np.newaxis, np.newaxis]\
        * sinchalf_Q[np.newaxis, np.newaxis, np.newaxis, :]\
        * sinchalf_CVdQ

    for Q_VdQv, test_Theta_VdQ, cell_Vcv in zip(Q_CVdQv, test_Theta_CVdQ,
                                                cell_CVcv):
        for Q_dQv, test_Theta_dQ, cell_cv in zip(Q_VdQv, test_Theta_VdQ,
                                                 cell_Vcv):
            for _ in range(3):  # Check that primary axis can be anywhere
                # Slide the cell axis indices
                cell_cv[:, :] = cell_cv[[2, 0, 1], :]
                # Calculate geometry factors
                Theta_dQ = parallelepipedic_geometry_factor(Q_dQv, cell_cv)

                # Check against expected results
                assert np.allclose(Theta_dQ, test_Theta_dQ, atol=1.e-8)


@pytest.mark.ci
def test_Co_hcp_site_kernels():
    """Check that the site kernel interface works on run-time inputs."""
    # ---------- Inputs ---------- #

    # Part 1: Generate plane wave representation (PWDescriptor)
    # Atomic configuration
    a = 2.5071
    c = 4.0695
    mm = 1.6
    # Ground state settings
    xc = 'LDA'
    kpts = 4
    pw = 200
    # Response settings
    ecut = 50.
    gammacentered = False
    q_c = [0., 0., 0.]
    qpm_qc = [[0., 0., 1 / 4.],
              [0., 0., -1. / 4.]]

    # Part 2: Calculate site kernels
    # Define partitions to try
    rc_pa = np.array([[1., 2.], [2., 3.]])  # radii in Å
    hc_pa = np.array([[2., 3.], [3., 4.]])  # heights in Å
    ez_pav = np.array([[[0., 0., 1.], [1., 0., 0.]],
                       [[0., 0., 1.], [0., 0., 1.]]])
    cell_acv = np.array([[[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]],
                         [[1., 0., 0.], [0., 1., 0.], [0., 0., 2.]]])
    cell_pacv = np.append(cell_acv, cell_acv * 2).reshape(2, 2, 3, 3)

    # Part 3: Check the calculated kernels

    # ---------- Script ---------- #

    # Part 1: Generate plane wave representation (PWDescriptor)
    atoms = bulk('Co', 'hcp', a=a, c=c)
    atoms.set_initial_magnetic_moments([mm, mm])

    calc = GPAW(xc=xc,
                spinpol=True,
                mode=PW(pw),
                kpts={'size': (kpts, kpts, kpts),
                      'gamma': True}
                )

    # Perform inexpensive calculator initialization
    calc.initialize(atoms)

    qpd0 = get_pw_descriptor(atoms, calc, q_c,
                             ecut=ecut,
                             gammacentered=gammacentered)

    # Part 2: Calculate site kernels
    positions = atoms.get_positions()

    # Generate spherical site kernels instances
    # Normally
    sph_sitekernels = SphericalSiteKernels(positions, rc_pa)
    # Separately as sum of site kernels
    sph_sitekernels0 = SphericalSiteKernels(positions[:1], rc_pa[:, :1])
    sph_sitekernels1 = SphericalSiteKernels(positions[1:], rc_pa[:, 1:])
    sph_sitekernels_sum = sph_sitekernels0 + sph_sitekernels1
    # Seperately as appended partitions
    sph_sitekernelsp0 = SphericalSiteKernels(positions, rc_pa[:1, :])
    sph_sitekernelsp1 = SphericalSiteKernels(positions, rc_pa[1:, :])
    sph_sitekernels_app = sph_sitekernelsp0.copy()
    sph_sitekernels_app.append(sph_sitekernelsp1)

    # Generate cylindrical site kernels instances
    # Normally
    cyl_sitekernels = CylindricalSiteKernels(positions, ez_pav, rc_pa, hc_pa)
    # Separately as a sum of site kernels
    cyl_sitekernels0 = CylindricalSiteKernels(positions[:1], ez_pav[:, :1, :],
                                              rc_pa[:, :1], hc_pa[:, :1])
    cyl_sitekernels1 = CylindricalSiteKernels(positions[1:], ez_pav[:, 1:, :],
                                              rc_pa[:, 1:], hc_pa[:, 1:])
    cyl_sitekernels_sum = cyl_sitekernels0 + cyl_sitekernels1
    # Seperately as appended partitions
    cyl_sitekernelsp0 = CylindricalSiteKernels(positions, ez_pav[:1, :, :],
                                               rc_pa[:1, :], hc_pa[:1, :])
    cyl_sitekernelsp1 = CylindricalSiteKernels(positions, ez_pav[1:, :, :],
                                               rc_pa[1:, :], hc_pa[1:, :])
    cyl_sitekernels_app = cyl_sitekernelsp0.copy()
    cyl_sitekernels_app.append(cyl_sitekernelsp1)

    # Generate parallelepipedic site kernels instances
    # Normally
    parlp_sitekernels = ParallelepipedicSiteKernels(positions, cell_pacv)
    # Separately as sum of site kernels
    parlp_sitekernels0 = ParallelepipedicSiteKernels(positions[:1],
                                                     cell_pacv[:, :1, ...])
    parlp_sitekernels1 = ParallelepipedicSiteKernels(positions[1:],
                                                     cell_pacv[:, 1:, ...])
    parlp_sitekernels_sum = parlp_sitekernels0 + parlp_sitekernels1
    # Seperately as appended partitions
    parlp_sitekernelsp0 = ParallelepipedicSiteKernels(positions,
                                                      cell_pacv[:1, :, ...])
    parlp_sitekernelsp1 = ParallelepipedicSiteKernels(positions,
                                                      cell_pacv[1:, :, ...])
    parlp_sitekernels_app = parlp_sitekernelsp0.copy()
    parlp_sitekernels_app.append(parlp_sitekernelsp1)

    # Collect all unique site kernels as both a sum and as different partitions
    # to show, that we can compute them in parallel
    all_sitekernels_sum = sph_sitekernels + cyl_sitekernels + parlp_sitekernels
    all_sitekernels_app = sph_sitekernels.copy()
    all_sitekernels_app.append(cyl_sitekernels)
    all_sitekernels_app.append(parlp_sitekernels)

    # Calculate spherical site kernels
    Ksph_paGG = np.array([K_aGG for K_aGG in
                          sph_sitekernels.calculate(qpd0)])
    Ksph0_paGG = np.array([K_aGG for K_aGG in
                           sph_sitekernels0.calculate(qpd0)])
    Ksph1_paGG = np.array([K_aGG for K_aGG in
                           sph_sitekernels1.calculate(qpd0)])
    Ksph_sum_paGG = np.array([K_aGG for K_aGG in
                              sph_sitekernels_sum.calculate(qpd0)])
    Ksphp0_paGG = np.array([K_aGG for K_aGG in
                            sph_sitekernelsp0.calculate(qpd0)])
    Ksphp1_paGG = np.array([K_aGG for K_aGG in
                            sph_sitekernelsp1.calculate(qpd0)])
    Ksph_app_paGG = np.array([K_aGG for K_aGG in
                              sph_sitekernels_app.calculate(qpd0)])

    # Calculate cylindrical site kernels
    Kcyl_paGG = np.array([K_aGG for K_aGG in
                          cyl_sitekernels.calculate(qpd0)])
    Kcyl0_paGG = np.array([K_aGG for K_aGG in
                           cyl_sitekernels0.calculate(qpd0)])
    Kcyl1_paGG = np.array([K_aGG for K_aGG in
                           cyl_sitekernels1.calculate(qpd0)])
    Kcyl_sum_paGG = np.array([K_aGG for K_aGG in
                              cyl_sitekernels_sum.calculate(qpd0)])
    Kcylp0_paGG = np.array([K_aGG for K_aGG in
                            cyl_sitekernelsp0.calculate(qpd0)])
    Kcylp1_paGG = np.array([K_aGG for K_aGG in
                            cyl_sitekernelsp1.calculate(qpd0)])
    Kcyl_app_paGG = np.array([K_aGG for K_aGG in
                              cyl_sitekernels_app.calculate(qpd0)])

    # Calculate parallelepipedic site kernels
    Kparlp_paGG = np.array([K_aGG for K_aGG in
                            parlp_sitekernels.calculate(qpd0)])
    Kparlp0_paGG = np.array([K_aGG for K_aGG in
                             parlp_sitekernels0.calculate(qpd0)])
    Kparlp1_paGG = np.array([K_aGG for K_aGG in
                             parlp_sitekernels1.calculate(qpd0)])
    Kparlp_sum_paGG = np.array([K_aGG for K_aGG in
                                parlp_sitekernels_sum.calculate(qpd0)])
    Kparlpp0_paGG = np.array([K_aGG for K_aGG in
                              parlp_sitekernelsp0.calculate(qpd0)])
    Kparlpp1_paGG = np.array([K_aGG for K_aGG in
                              parlp_sitekernelsp1.calculate(qpd0)])
    Kparlp_app_paGG = np.array([K_aGG for K_aGG in
                                parlp_sitekernels_app.calculate(qpd0)])

    # Calculate all site kernels together
    Kall_sum_paGG = np.array([K_aGG for K_aGG in
                              all_sitekernels_sum.calculate(qpd0)])
    Kall_app_paGG = np.array([K_aGG for K_aGG in
                              all_sitekernels_app.calculate(qpd0)])

    # Calculate all site kernels at opposite qs
    qpd_q = [get_pw_descriptor(atoms, calc, qpm_c,
                               ecut=ecut,
                               gammacentered=gammacentered)
             for qpm_c in qpm_qc]
    Kall_pm_qpaGG = [np.array([K_aGG for K_aGG in
                               all_sitekernels_app.calculate(qpd)])
                     for qpd in qpd_q]
    
    # Part 4: Check the calculated kernels

    # Check geometry shapes of basic arrays
    assert all([gs == 'sphere' for gs in sph_sitekernels.geometry_shapes])
    assert all([gs == 'cylinder' for gs in cyl_sitekernels.geometry_shapes])
    assert all([gs == 'parallelepiped'
                for gs in parlp_sitekernels.geometry_shapes])

    # Check geometry shapes of summed arrays
    assert all([gs == 'sphere' for gs in sph_sitekernels_sum.geometry_shapes])
    assert all([gs == 'cylinder'
                for gs in cyl_sitekernels_sum.geometry_shapes])
    assert all([gs == 'parallelepiped'
                for gs in parlp_sitekernels_sum.geometry_shapes])
    assert all([gs is None for gs in all_sitekernels_sum.geometry_shapes])

    # Check geometry shapes of appended arrays
    assert all([gs == 'sphere' for gs in sph_sitekernels_app.geometry_shapes])
    assert all([gs == 'cylinder'
                for gs in cyl_sitekernels_app.geometry_shapes])
    assert all([gs == 'parallelepiped'
                for gs in parlp_sitekernels_app.geometry_shapes])
    gs_refs = 2 * ['sphere'] + 2 * ['cylinder'] + 2 * ['parallelepiped']
    assert all([gs == ref for gs, ref in
                zip(all_sitekernels_app.geometry_shapes, gs_refs)])

    # Check shape of spherical kernel arrays
    nG = len(get_pw_coordinates(qpd0))
    assert sph_sitekernels.shape == Ksph_paGG.shape[:2]
    assert Ksph_paGG.shape == rc_pa.shape + (nG, nG)
    assert Ksph0_paGG.shape == (rc_pa.shape[0], 1) + (nG, nG)
    assert Ksph0_paGG.shape == Ksph1_paGG.shape
    assert Ksph_sum_paGG.shape == rc_pa.shape + (nG, nG)
    assert Ksphp0_paGG.shape == (1, rc_pa.shape[1]) + (nG, nG)
    assert Ksphp0_paGG.shape == Ksphp1_paGG.shape
    assert Ksph_app_paGG.shape == rc_pa.shape + (nG, nG)

    # Check shape of cylindrical kernel arrays
    assert cyl_sitekernels.shape == Kcyl_paGG.shape[:2]
    assert Kcyl_paGG.shape == rc_pa.shape + (nG, nG)
    assert Kcyl0_paGG.shape == (rc_pa.shape[0], 1) + (nG, nG)
    assert Kcyl0_paGG.shape == Kcyl1_paGG.shape
    assert Kcyl_sum_paGG.shape == rc_pa.shape + (nG, nG)
    assert Kcylp0_paGG.shape == (1, rc_pa.shape[1]) + (nG, nG)
    assert Kcylp0_paGG.shape == Kcylp1_paGG.shape
    assert Kcyl_app_paGG.shape == rc_pa.shape + (nG, nG)

    # Check shape of parallelepipedic kernel arrays
    assert parlp_sitekernels.shape == Kparlp_paGG.shape[:2]
    assert Kparlp_paGG.shape == cell_pacv.shape[:2] + (nG, nG)
    assert Kparlp0_paGG.shape == (cell_pacv.shape[0], 1) + (nG, nG)
    assert Kparlp0_paGG.shape == Kparlp1_paGG.shape
    assert Kparlp_sum_paGG.shape == cell_pacv.shape[:2] + (nG, nG)
    assert Kparlpp0_paGG.shape == (1, cell_pacv.shape[1]) + (nG, nG)
    assert Kparlpp0_paGG.shape == Kparlpp1_paGG.shape
    assert Kparlp_app_paGG.shape == cell_pacv.shape[:2] + (nG, nG)

    # Check shape of array calculated in parallel
    assert Kall_sum_paGG.shape == (2, 6, nG, nG)
    assert Kall_app_paGG.shape == (6, 2, nG, nG)

    # Check self-consitency of spherical arrays
    assert np.allclose(Ksph0_paGG[:, 0, ...], Ksph_sum_paGG[:, 0, ...])
    assert np.allclose(Ksph1_paGG[:, 0, ...], Ksph_sum_paGG[:, 1, ...])
    assert np.allclose(Ksph_paGG, Ksph_sum_paGG)
    assert np.allclose(Ksphp0_paGG[0], Ksph_app_paGG[0])
    assert np.allclose(Ksphp1_paGG[0], Ksph_app_paGG[1])
    assert np.allclose(Ksph_paGG, Ksph_app_paGG)

    # Check self-consistency of cylindrical arrays
    assert np.allclose(Kcyl0_paGG[:, 0, ...], Kcyl_sum_paGG[:, 0, ...])
    assert np.allclose(Kcyl1_paGG[:, 0, ...], Kcyl_sum_paGG[:, 1, ...])
    assert np.allclose(Kcyl_paGG, Kcyl_sum_paGG)
    assert np.allclose(Kcylp0_paGG[0], Kcyl_app_paGG[0])
    assert np.allclose(Kcylp1_paGG[0], Kcyl_app_paGG[1])
    assert np.allclose(Kcyl_paGG, Kcyl_app_paGG)

    # Check self-consistency of parallelepipedic arrays
    assert np.allclose(Kparlp0_paGG[:, 0, ...], Kparlp_sum_paGG[:, 0, ...])
    assert np.allclose(Kparlp1_paGG[:, 0, ...], Kparlp_sum_paGG[:, 1, ...])
    assert np.allclose(Kparlp_paGG, Kparlp_sum_paGG)
    assert np.allclose(Kparlpp0_paGG[0], Kparlp_app_paGG[0])
    assert np.allclose(Kparlpp1_paGG[0], Kparlp_app_paGG[1])
    assert np.allclose(Kparlp_paGG, Kparlp_app_paGG)

    # Check self-consistency of kernels calculated in parallel
    assert np.allclose(Ksph_paGG, Kall_sum_paGG[:, :2])
    assert np.allclose(Kcyl_paGG, Kall_sum_paGG[:, 2:4])
    assert np.allclose(Kparlp_paGG, Kall_sum_paGG[:, -2:])
    assert np.allclose(Ksph_paGG, Kall_app_paGG[:2])
    assert np.allclose(Kcyl_paGG, Kall_app_paGG[2:4])
    assert np.allclose(Kparlp_paGG, Kall_app_paGG[-2:])

    # Check that K_G=G'(q=0) gives Vint / V0 (fractional integration volume)
    # Volume of unit cell in Å^3
    V0 = atoms.get_volume()
    # Calculate integration volumes in Å^3
    Vsphere_pa = 4 / 3 * np.pi * rc_pa**3
    Vcylinder_pa = np.pi * rc_pa**2 * hc_pa
    Vparlp_pa = np.abs(np.linalg.det(cell_pacv))
    assert np.allclose(np.diagonal(Ksph_paGG, axis1=2, axis2=3),
                       Vsphere_pa[..., np.newaxis] / V0)
    assert np.allclose(np.diagonal(Kcyl_paGG, axis1=2, axis2=3),
                       Vcylinder_pa[..., np.newaxis] / V0)
    assert np.allclose(np.diagonal(Kparlp_paGG, axis1=2, axis2=3),
                       Vparlp_pa[..., np.newaxis] / V0)

    # Check that K_G=G'(q) gives e^(-iq.τ_a) Θ(q) / V0
    B_cv = 2.0 * np.pi * atoms.cell.reciprocal()  # Coordinate transform
    for qpm_c, Kall_pm_paGG in zip(qpm_qc, Kall_pm_qpaGG):
        # Calculate q-vector in Å^(-1)
        qpm_v = qpm_c @ B_cv
        exp_a = np.exp(-1.j * positions @ qpm_v)

        # Calculate site centered geometry factors
        Theta_pa = []
        # Spherical geometry factors
        for rc_a in rc_pa:
            Theta_a = []
            for rc in rc_a:
                Theta_a.append(spherical_geometry_factor(qpm_v, rc))
            Theta_pa.append(Theta_a)
        # Cylindrical geometry factors
        for ez_av, rc_a, hc_a in zip(ez_pav, rc_pa, hc_pa):
            Theta_a = []
            for ez_v, rc, hc in zip(ez_av, rc_a, hc_a):
                Theta_a.append(cylindrical_geometry_factor(qpm_v, ez_v,
                                                           rc, hc))
            Theta_pa.append(Theta_a)
        # Parallelepipedic geometry factors
        for cell_acv in cell_pacv:
            Theta_a = []
            for cell_cv in cell_acv:
                Theta_a.append(parallelepipedic_geometry_factor(qpm_v,
                                                                cell_cv))
            Theta_pa.append(Theta_a)
        Theta_pa = np.array(Theta_pa)

        assert np.allclose(np.diagonal(Kall_pm_paGG, axis1=2, axis2=3),
                           Theta_pa[..., np.newaxis]
                           * exp_a[np.newaxis, :, np.newaxis] / V0)

    # Check that K_GG'(q) is the hermitian conjugate of K_GG'(-q)
    Kall_mH_paGG = np.conjugate(np.transpose(Kall_pm_qpaGG[1], (0, 1, 3, 2)))
    assert np.allclose(Kall_pm_qpaGG[0], Kall_mH_paGG)


# ---------- Test functionality ---------- #


def get_pw_descriptor(atoms, calc, q_c, ecut=50., gammacentered=False):
    """Mock-up of ChiKSCalculator.get_pw_descriptor.

    Works on a bare calculator instance without any actual data in it."""
    from ase.units import Ha
    from gpaw.response.pair_functions import SingleQPWDescriptor

    # Create the plane wave descriptor
    q_c = np.asarray(q_c, dtype=float)
    qpd = SingleQPWDescriptor.from_q(q_c, ecut / Ha, calc.wfs.gd,
                                     gammacentered=gammacentered)
    return qpd



import pytest
import numpy as np
from gpaw.grid_descriptor import GridDescriptor
from gpaw.test import equal


def test_fd_ops_gd():
    gd = GridDescriptor([4, 4, 4])
    a = gd.empty(dtype=complex)
    a[:] = 1.0
    assert gd.integrate(a.real, a.real) == 1.0

    def mic(r_v, pbc_c):
        "Mnimal image convention in an [1.0] * 3 unit cell"
        r_c = r_v % 1.0
        r_c -= (2 * r_c).astype(int)
        return np.where(pbc_c, r_c, r_v)

    pbc_c = [1, 1, 0]
    gd = GridDescriptor([10] * 3, cell_cv=[1.] * 3, pbc_c=pbc_c)
    # Point outside box in non-periodic direction should stay
    r_v = np.array([0.01, 2.49, 0.01])
    dr_cG = gd.get_grid_point_distance_vectors(r_v)
    equal(dr_cG[:, 0, 0, 0], mic(
        np.dot(gd.h_cv, gd.beg_c) - r_v, pbc_c), 1e-15)
    # Point outside box in periodic direction should be folded inside
    r_v = np.array([2.49, 0.01, 0.01])
    dr_cG = gd.get_grid_point_distance_vectors(r_v)
    equal(dr_cG[:, 0, 0, 0], mic(
        np.dot(gd.h_cv, gd.beg_c) - r_v, pbc_c), 1e-15)


@pytest.mark.serial
def test_nonorthogonal_distances():
    """test distance vectors in non-orthogonal cells"""
    cell_cv = np.eye(3)
    cell_cv[0, 1] = 1

    gd = GridDescriptor([2, 4, 1], cell_cv)
    r_vG = gd.get_grid_point_distance_vectors((0, 0, 0))

    assert r_vG[1, 1].reshape(4) == pytest.approx([-0.5, -0.25, -1., -0.75])



"""Test Hirshfeld for spin/no spin consistency."""
import pytest
from ase import Atom
from ase.parallel import parprint

from gpaw import GPAW, FermiDirac
from gpaw.analyse.hirshfeld import HirshfeldPartitioning
from gpaw.cluster import Cluster
from gpaw.test import equal


@pytest.mark.later
def test_vdw_H_Hirshfeld():
    h = 0.25
    box = 3

    atoms = Cluster()
    atoms.append(Atom('H'))
    atoms.minimal_box(box)

    volumes = []
    for spinpol in [False, True]:
        calc = GPAW(h=h,
                    occupations=FermiDirac(0.1, fixmagmom=spinpol),
                    experimental={'niter_fixdensity': 2},
                    spinpol=spinpol)
        calc.calculate(atoms)
        vol = HirshfeldPartitioning(calc).get_effective_volume_ratios()
        volumes.append(vol)
    parprint(volumes)
    equal(volumes[0][0], volumes[1][0], 4e-9)



from math import pi

import numpy as np
import pytest

from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
from gpaw.grid_descriptor import GridDescriptor
from gpaw.xc import XC
from gpaw.test import equal


@pytest.mark.ci
def test_xc_XC2():
    rgd = EquidistantRadialGridDescriptor(0.01, 100)

    for name in ['LDA', 'PBE']:
        xc = XC(name)
        for nspins in [1, 2]:
            n = rgd.zeros(nspins)
            v = rgd.zeros(nspins)
            n[:] = np.exp(-rgd.r_g**2)
            n[-1] *= 2
            E = xc.calculate_spherical(rgd, n, v)
            i = 23
            x = v[-1, i] * rgd.dv_g[i]
            n[-1, i] += 0.000001
            Ep = xc.calculate_spherical(rgd, n, v)
            n[-1, i] -= 0.000002
            Em = xc.calculate_spherical(rgd, n, v)
            x2 = (Ep - Em) / 0.000002
            print(name, nspins, E, x, x2, x - x2)
            equal(x, x2, 1e-9)
            n[-1, i] += 0.000001
            if nspins == 1:
                ns = rgd.empty(2)
                ns[:] = n / 2
                Es = xc.calculate_spherical(rgd, ns, 0 * ns)
                equal(E, Es, 1e-13)

    N = 20
    a = 1.0
    gd = GridDescriptor((N, N, N), (a, a, a))

    for name in ['LDA', 'PBE']:
        xc = XC(name)
        for nspins in [1, 2]:
            n = gd.empty(nspins)
            n.fill(0.03)
            z = np.arange(gd.beg_c[2], gd.end_c[2]) * a / N
            n[:] += 0.01 * np.sin(2 * pi * z / a)
            if nspins == 2:
                n[1] += 0.01 * np.cos(2 * pi * z / a)
            n /= nspins

            v = 0.0 * n
            E = xc.calculate(gd, n, v)

            here = (gd.beg_c[0] <= 1 < gd.end_c[0] and
                    gd.beg_c[1] <= 2 < gd.end_c[1] and
                    gd.beg_c[2] <= 3 < gd.end_c[2])
            if here:
                x = v[-1, 1, 2, 3] * gd.dv
                n[-1, 1, 2, 3] += 0.000001
            Ep = xc.calculate(gd, n, v)
            if here:
                n[-1, 1, 2, 3] -= 0.000002
            Em = xc.calculate(gd, n, v)
            x2 = (Ep - Em) / 0.000002
            if here:
                print(name, nspins, E, x, x2, x - x2)
                equal(x, x2, 1e-11)
                n[-1, 1, 2, 3] += 0.000001

            if nspins == 1:
                ns = gd.empty(2)
                ns[:] = n / 2
                Es = xc.calculate(gd, ns, 0 * ns)
                equal(E, Es, 1e-13)



from gpaw.cluster import Cluster
from ase.build import molecule
from ase.data.vdw import vdw_radii
from gpaw.solvation import (
    SolvationGPAW,
    EffectivePotentialCavity,
    Power12Potential,
    LinearDielectric)
from gpaw.solvation.poisson import ADM12PoissonSolver
import warnings

h = 0.3
vac = 3.0
u0 = .180
epsinf = 80.
T = 298.15
vdw_radii = vdw_radii.copy()
vdw_radii[1] = 1.09


def atomic_radii(atoms):
    return [vdw_radii[n] for n in atoms.numbers]


convergence = {
    'energy': 0.05 / 8.,
    'density': 10.,
    'eigenstates': 10.,
}


def test_solvation_pbc():
    atoms = Cluster(molecule('H2O'))
    atoms.minimal_box(vac, h)
    atoms.pbc = True

    with warnings.catch_warnings():
        # ignore production code warning for ADM12PoissonSolver
        warnings.simplefilter("ignore")
        psolver = ADM12PoissonSolver(eps=1e-7)

    atoms.calc = SolvationGPAW(
        xc='LDA', h=h, convergence=convergence,
        cavity=EffectivePotentialCavity(
            effective_potential=Power12Potential(
                atomic_radii=atomic_radii, u0=u0),
            temperature=T
        ),
        dielectric=LinearDielectric(epsinf=epsinf),
        poissonsolver=psolver
    )
    atoms.get_potential_energy()
    atoms.get_forces()



import numpy as np
from numpy import sqrt, pi, exp, abs
from scipy.special import erf

import _gpaw
from gpaw import debug
from gpaw.utilities.tools import coordinates
from gpaw.utilities import is_contiguous


def Y_L(L, x, y, z, r2):
    if L == 0:
        return 0.28209479177387814
    elif L == 1:
        return 0.48860251190291992 * y
    elif L == 2:
        return 0.48860251190291992 * z
    elif L == 3:
        return 0.48860251190291992 * x
    elif L == 4:
        return 1.0925484305920792 * x * y
    elif L == 5:
        return 1.0925484305920792 * y * z
    elif L == 6:
        return 0.31539156525252005 * (3 * z * z - r2)
    elif L == 7:
        return 1.0925484305920792 * x * z
    elif L == 8:
        return 0.54627421529603959 * (x * x - y * y)


def gauss_L(a, L, x, y, z, r2, exp_ar2):
    if L == 0:
        return sqrt(a**3 * 4) / pi * exp_ar2
    elif L == 1:
        return sqrt(a**5 * 5.333333333333333) / pi * y * exp_ar2
    elif L == 2:
        return sqrt(a**5 * 5.333333333333333) / pi * z * exp_ar2
    elif L == 3:
        return sqrt(a**5 * 5.333333333333333) / pi * x * exp_ar2
    elif L == 4:
        return sqrt(a**7 * 4.2666666666666666) / pi * x * y * exp_ar2
    elif L == 5:
        return sqrt(a**7 * 4.2666666666666666) / pi * y * z * exp_ar2
    elif L == 6:
        return sqrt(
            a**7 * 0.35555555555555557) / pi * (3 * z * z - r2) * exp_ar2
    elif L == 7:
        return sqrt(a**7 * 4.2666666666666666) / pi * x * z * exp_ar2
    elif L == 8:
        return sqrt(a**7 * 1.0666666666666667) / pi * (x * x - y * y) * exp_ar2


def gausspot_L(a, L, x, y, z, r, r2, erf_sar, exp_ar2):
    if L == 0:
        return 2.0 * 1.7724538509055159 * erf_sar / r
    elif L == 1:
        return 1.1547005383792515 * (1.7724538509055159 * erf_sar -
                                     2 * sqrt(a) * r * exp_ar2) / r / r2 * y
    elif L == 2:
        return 1.1547005383792515 * (1.7724538509055159 * erf_sar -
                                     2 * sqrt(a) * r * exp_ar2) / r / r2 * z
    elif L == 3:
        return 1.1547005383792515 * (1.7724538509055159 * erf_sar -
                                     2 * sqrt(a) * r * exp_ar2) / r / r2 * x
    elif L == 4:
        return 0.5163977794943222 * (
            5.3173615527165481 * erf_sar -
            (6 + 4 *
             (sqrt(a) * r)**2) * sqrt(a) * r * exp_ar2) / r / r2**2 * x * y
    elif L == 5:
        return 0.5163977794943222 * (
            5.3173615527165481 * erf_sar -
            (6 + 4 *
             (sqrt(a) * r)**2) * sqrt(a) * r * exp_ar2) / r / r2**2 * y * z
    elif L == 6:
        return 0.14907119849998599 * (5.3173615527165481 * erf_sar -
                                      (6 + 4 *
                                       (sqrt(a) * r)**2) * sqrt(a) * r *
                                      exp_ar2) / r / r2**2 * (3 * z * z - r2)
    elif L == 7:
        return 0.5163977794943222 * (
            5.3173615527165481 * erf_sar -
            (6 + 4 *
             (sqrt(a) * r)**2) * sqrt(a) * r * exp_ar2) / r / r2**2 * x * z
    elif L == 8:
        return 0.2581988897471611 * (5.3173615527165481 * erf_sar -
                                     (6 + 4 * (sqrt(a) * r)**2) * sqrt(a) * r *
                                     exp_ar2) / r / r2**2 * (x * x - y * y)


# end of computer generated code


class Gaussian:
    r"""Class offering several utilities related to the generalized gaussians.

    Generalized gaussians are defined by::

                       _____                           2
                      /  1       l!         l+3/2  -a r   l  m
       g (x,y,z) =   / ----- --------- (4 a)      e      r  Y (x,y,z),
        L          \/  4 pi  (2l + 1)!                       l

    where a is the inverse width of the gaussian, and Y_l^m is a real
    spherical harmonic.
    The gaussians are centered in the middle of input grid-descriptor."""
    def __init__(self, gd, a=19., center=None):
        self.gd = gd
        self.xyz, self.r2 = coordinates(gd, center)
        self.r = np.sqrt(self.r2)
        self.set_width(a, center)
        self.exp_ar2 = exp(-self.a * self.r2)
        self.erf_sar = erf(sqrt(self.a) * self.r)

    def set_width(self, a, center):
        """Set exponent of exp-function to -a on the boundary."""
        if center is None:
            self.a = 4 * a * (self.gd.icell_cv**2).sum(1).max()
        else:
            cell_center = self.gd.cell_cv.sum(1) / 2
            r_min = (cell_center - np.abs(center - cell_center)).min()
            self.a = a / r_min**2

    def get_gauss(self, L):
        a = self.a
        x, y, z = tuple(self.xyz)
        r2 = self.r2
        exp_ar2 = self.exp_ar2
        return gauss_L(a, L, x, y, z, r2, exp_ar2)

    def get_gauss_pot(self, L):
        a = self.a
        x, y, z = tuple(self.xyz)
        r2 = self.r2
        r = self.r
        erf_sar = self.erf_sar
        exp_ar2 = self.exp_ar2
        return gausspot_L(a, L, x, y, z, r, r2, erf_sar, exp_ar2)

    def get_moment(self, n, L):
        r2 = self.r2
        x, y, z = tuple(self.xyz)
        return self.gd.integrate(n * Y_L(L, x, y, z, r2))

    def remove_moment(self, n, L, q=None):
        # Determine multipole moment
        if q is None:
            q = self.get_moment(n, L)

        # Don't do anything if moment is less than the tolerance
        if abs(q) < 1e-7:
            return 0.

        # Remove moment from input density
        n -= q * self.get_gauss(L)

        # Return correction
        return q * self.get_gauss_pot(L)


def gaussian_wave(r_vG, r0_v, sigma, k_v=None, A=None, dtype=float,
                  out_G=None):
    r"""Generates function values for atom-centered Gaussian waves.

    ::

                         _ _
        _            / -|r-r0|^2 \           _ _
      f(r) = A * exp( ----------- ) * exp( i k.r )
                     \ 2 sigma^2 /

    If the parameter A is not specified, the Gaussian wave is normalized::

                                                  oo
           /    ____        \ -3/2               /       _  2  2
      A = (    /    '        )        =>    4 pi | dr |f(r)|  r  = 1
           \ \/  pi   sigma /                    /
                                                   0

    Parameters:

    r_vG: ndarray
        Set of coordinates defining the grid positions.
    r0_v: ndarray
        Set of coordinates defining the center of the Gaussian envelope.
    sigma: float
        Specifies the spatial width of the Gaussian envelope.
    k_v: ndarray or None
        Set of reciprocal lattice coordinates defining the wave vector.
        An argument of None is interpreted as the gamma point i.e. k_v=0.
    A: float, complex or None
        Specifies the amplitude of the Gaussian wave. Normalizes if None.
    dtype: type, defaults to float
        Specifies the output data type. Only returns the real-part if float.
    out_G: ndarray or None
        Optional pre-allocated buffer to fill in values. Allocates if None.

    """
    if k_v is None:
        k_v = np.zeros(r0_v.shape)

    if A is None:
        # 4*pi*int(exp(-r^2/(2*sigma^2))^2 * r^2, r=0...infinity)
        # = sigma^3*pi^(3/2) = 1/A^2 -> A = (sqrt(Pi)*sigma)^(-3/2)
        A = 1 / (sigma * np.pi**0.5)**1.5

    if debug:
        assert is_contiguous(r_vG, float)
        assert is_contiguous(r0_v, float)
        assert is_contiguous(k_v, float)
        assert r_vG.ndim >= 2 and r_vG.shape[0] > 0
        assert r0_v.ndim == 1 and r0_v.shape[0] > 0
        assert k_v.ndim == 1 and k_v.shape[0] > 0
        assert (r_vG.shape[0], ) == r0_v.shape == k_v.shape
        assert sigma > 0

    if out_G is None:
        out_G = np.empty(r_vG.shape[1:], dtype=dtype)
    elif debug:
        assert is_contiguous(out_G)
        assert out_G.shape == r_vG.shape[1:]

    # slice_v2vG = [slice(None)] + [np.newaxis]*3
    # gw = lambda r_vG, r0_v, sigma, k_v, A=1/(sigma*np.pi**0.5)**1.5: \
    #    * np.exp(-np.sum((r_vG-r0_v[slice_v2vG])**2, axis=0)/(2*sigma**2)) \
    #    * np.exp(1j*np.sum(np.r_vG*k_v[slice_v2vG], axis=0)) * A
    _gpaw.utilities_gaussian_wave(A, r_vG, r0_v, sigma, k_v, out_G)
    return out_G
